#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.util import tentacle_fetching
from octobot_tentacles_manager.util import tentacle_explorer
from octobot_tentacles_manager.util import tentacle_filter
from octobot_tentacles_manager.util import tentacle_cleaner
from octobot_tentacles_manager.util import file_util

from octobot_tentacles_manager.util.tentacle_fetching import (
    cleanup_temp_dirs,
    fetch_and_extract_tentacles,
    get_local_arch_download_path,
)
from octobot_tentacles_manager.util.tentacle_explorer import (
    load_tentacle_with_metadata,
    get_tentacles_from_package,
)
from octobot_tentacles_manager.util.file_util import (
    find_or_create,
    replace_with_remove_or_rename,
    merge_folders,
)
from octobot_tentacles_manager.util.tentacle_filter import (
    TentacleFilter,
    filter_tentacles_by_dev_mode_and_package,
)
from octobot_tentacles_manager.util.tentacle_cleaner import (
    remove_unnecessary_files,
    remove_non_tentacles_files,
    remove_dir_or_file_from_path,
    remove_dir_or_file,
)

__all__ = [
    "cleanup_temp_dirs",
    "remove_dir_or_file",
    "remove_dir_or_file_from_path",
    "fetch_and_extract_tentacles",
    "get_local_arch_download_path",
    "load_tentacle_with_metadata",
    "get_tentacles_from_package",
    "find_or_create",
    "get_tentacles_from_package",
    "replace_with_remove_or_rename",
    "merge_folders",
    "TentacleFilter",
    "remove_unnecessary_files",
    "remove_non_tentacles_files",
    "filter_tentacles_by_dev_mode_and_package",
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_tentacles_manager.models.artifact as artifact


class Profile(artifact.Artifact):
    ARTIFACT_NAME = "profile"

    def __init__(self, name):
        super().__init__(name)

    def __str__(self):
        str_rep: str = f"{self.name} {Profile.ARTIFACT_NAME} ["
        if self.is_valid():
            return f"{str_rep}version: {self.version}]"
        else:
            return f"{str_rep}]"



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import os
import shutil
import zipfile

import aiofiles
import aiohttp
import time

import pytest

import octobot_tentacles_manager.api.uploader as uploader_api
import octobot_tentacles_manager.uploaders as uploaders
import octobot_tentacles_manager.constants as constants

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

TEST_NEXUS_DIRECTORY = "nexus-tests"
TEST_NEXUS_PATH = "officials/base/dev/tests/tentacle-manager/"
TEST_NEXUS_FILE_NAME = "test"


@pytest.yield_fixture()
async def nexus_tests():
    if os.path.exists(TEST_NEXUS_DIRECTORY):
        shutil.rmtree(TEST_NEXUS_DIRECTORY)
    os.mkdir(TEST_NEXUS_DIRECTORY)
    yield
    if os.path.exists(TEST_NEXUS_DIRECTORY):
        shutil.rmtree(TEST_NEXUS_DIRECTORY)


async def test_upload_file(nexus_tests):
    aiohttp_session = aiohttp.ClientSession()
    nexus_test_file_name: str = f"{time.time_ns()}.json"
    local_file_name: str = os.path.join(TEST_NEXUS_DIRECTORY, f"{TEST_NEXUS_FILE_NAME}.json")

    # test upload file
    with open(local_file_name, "w") as test_file:
        test_file.write(json.dumps({'test-key': 1}))
    assert await uploader_api.upload_file_or_folder_to_nexus(nexus_path=TEST_NEXUS_PATH,
                                                             artifact_path=local_file_name,
                                                             artifact_alias=nexus_test_file_name) == 0
    # test download file
    downloaded_file_path: str = await download_file_from_nexus(aiohttp_session,
                                                               f"{TEST_NEXUS_PATH}{nexus_test_file_name}",
                                                               "downloaded_file")

    with open(downloaded_file_path, "r") as downloaded_file:
        assert json.loads(downloaded_file.read()) == {
            'test-key': 1
        }
    await aiohttp_session.close()


async def test_upload_folder(nexus_tests):
    aiohttp_session = aiohttp.ClientSession()
    test_dir_path = os.path.join(TEST_NEXUS_DIRECTORY, "test-dir")
    os.mkdir(test_dir_path)

    nexus_test_file_name: str = f"{time.time_ns()}"
    local_file_name: str = os.path.join(test_dir_path, f"{TEST_NEXUS_FILE_NAME}.json")
    local_zip_path: str = os.path.join(test_dir_path, f"{TEST_NEXUS_FILE_NAME}")

    # test upload file
    with open(local_file_name, "w") as test_file:
        test_file.write(json.dumps({'test-key': 1}))
    shutil.make_archive(local_zip_path, constants.TENTACLES_PACKAGE_FORMAT, test_dir_path)
    assert await uploader_api.upload_file_or_folder_to_nexus(nexus_path=TEST_NEXUS_PATH,
                                                             artifact_path=test_dir_path,
                                                             artifact_alias=nexus_test_file_name) == 0
    # test download folder files
    downloaded_file_path: str = await download_file_from_nexus(aiohttp_session,
                                                               f"{TEST_NEXUS_PATH}{nexus_test_file_name}/test.json",
                                                               "downloaded_test.json")
    downloaded_zip_path: str = await download_file_from_nexus(aiohttp_session,
                                                              f"{TEST_NEXUS_PATH}{nexus_test_file_name}/test.zip",
                                                              "downloaded_test.zip")
    with open(downloaded_file_path, "r") as downloaded_file:
        assert json.loads(downloaded_file.read()) == {
            'test-key': 1
        }
    zipfile.is_zipfile(downloaded_zip_path)
    await aiohttp_session.close()


async def download_file_from_nexus(aiohttp_session: aiohttp.ClientSession,
                                   file_url: str,
                                   local_file_name: str) -> str:
    downloaded_file_path: str = os.path.join(TEST_NEXUS_DIRECTORY, local_file_name)
    resp = await aiohttp_session.get(f"{os.getenv(uploaders.NexusUploader.ENV_NEXUS_URL)}/{file_url}")
    assert resp.status == 200
    async with aiofiles.open(downloaded_file_path, mode='wb') as downloaded_file:
        await downloaded_file.write(await resp.read())
    return downloaded_file_path



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os.path as path

import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.models as models
import octobot_tentacles_manager.util as util

# tentacle_data_by_tentacle_class is used to cache tentacles metadata
_tentacle_by_tentacle_class = None


def reload_tentacle_by_tentacle_class(tentacles_path=constants.TENTACLES_PATH):
    global _tentacle_by_tentacle_class
    loaded_tentacles = util.load_tentacle_with_metadata(tentacles_path)
    _tentacle_by_tentacle_class = {
        klass: tentacle
        for tentacle in loaded_tentacles
        for klass in tentacle.tentacle_class_names
    }


def get_tentacle_classes() -> dict:
    return _tentacle_by_tentacle_class


def ensure_tentacles_metadata(tentacles_path) -> None:
    if get_tentacle_classes() is None:
        reload_tentacle_by_tentacle_class(tentacles_path)


def get_resources_path(klass) -> str:
    return path.join(get_tentacle_module_path(klass), constants.TENTACLE_RESOURCES)


def get_tentacle_module_path(klass) -> str:
    tentacle = get_tentacle(klass)
    return path.join(tentacle.tentacle_path, tentacle.name)


def get_documentation_file_path(klass) -> str:
    return \
        path.join(get_resources_path(klass), f"{klass if isinstance(klass, str) else klass.get_name()}"
                                             f"{constants.DOCUMENTATION_EXT}")


def get_tentacle(klass) -> models.Tentacle:
    try:
        return _tentacle_by_tentacle_class[klass if isinstance(klass, str) else klass.get_name()]
    except TypeError:
        raise RuntimeError(f"tentacle have not been initialized, call reload_tentacle_data_by_tentacle_class "
                           f"fix this issue")



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_tentacles_manager.enums as enums
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.models.artifact as artifact_model


class ArtifactMetadata:
    def __init__(self, artifact: artifact_model.Artifact):
        self.artifact: artifact_model.Artifact = artifact
        self.version: str = self.artifact.version
        self.author: str = constants.DEFAULT_ARTIFACT_METADATA_AUTHOR
        self.artifact_type: enums.ArtifactTypes = None

    def to_dict(self) -> dict:
        return {
            constants.ARTIFACT_METADATA_NAME: self.artifact.name,
            constants.ARTIFACT_METADATA_VERSION: self.artifact.version,
            constants.ARTIFACT_METADATA_ARTIFACT_TYPE: self.artifact_type.value,
            constants.ARTIFACT_METADATA_REPOSITORY: self.artifact.origin_repository,
            constants.ARTIFACT_METADATA_AUTHOR: self.author,
        }



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.managers import tentacle_manager
from octobot_tentacles_manager.managers import tentacles_setup_manager
from octobot_tentacles_manager.managers import tentacles_init_files_manager
from octobot_tentacles_manager.managers import python_modules_requirements_manager
from octobot_tentacles_manager.managers import profiles_manager

from octobot_tentacles_manager.managers.tentacle_manager import (
    TentacleManager,
)
from octobot_tentacles_manager.managers.tentacles_setup_manager import (
    TentaclesSetupManager,
)
from octobot_tentacles_manager.managers.tentacles_init_files_manager import (
    update_tentacle_type_init_file,
    get_module_init_file_content,
    get_tentacle_import_block,
    find_or_create_module_init_file,
    create_tentacle_init_file_if_necessary,
)
from octobot_tentacles_manager.managers.python_modules_requirements_manager import (
    install_tentacle,
    update_tentacle,
    list_installed_tentacles,
)
from octobot_tentacles_manager.managers.profiles_manager import (
    get_profile_folders,
    import_profile,
)

__all__ = [
    "TentacleManager",
    "TentaclesSetupManager",
    "update_tentacle_type_init_file",
    "get_module_init_file_content",
    "get_tentacle_import_block",
    "find_or_create_module_init_file",
    "create_tentacle_init_file_if_necessary",
    "install_tentacle",
    "update_tentacle",
    "list_installed_tentacles",
    "get_profile_folders",
    "import_profile",
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import os.path as path

import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.util.tentacle_explorer as explorer


def filter_tentacles_by_dev_mode_and_package(tentacles: list,
                                             with_dev_mode: bool = False,
                                             package_filter: str = None) -> list:
    # remove dev-mode tentacles if necessary
    tentacles_white_list = tentacles if with_dev_mode else _filter_in_dev_tentacles(tentacles)
    if package_filter is not None:
        # only keep tentacles from the tentacles package to export
        tentacles_white_list = explorer.get_tentacles_from_package(tentacles_white_list, package_filter)
    return tentacles_white_list


def _filter_in_dev_tentacles(tentacles):
    return [
        tentacle
        for tentacle in tentacles
        if not tentacle.in_dev_mode
    ]


class TentacleFilter:
    def __init__(self, full_tentacles_list, tentacles_white_list):
        self.tentacles_white_list = tentacles_white_list
        self.full_tentacles_list = full_tentacles_list
        self.tentacle_paths_black_list = [] if self.tentacles_white_list is None else [
            path.join(tentacle.tentacle_path, tentacle.name)
            for tentacle in self.full_tentacles_list
            if tentacle not in self.tentacles_white_list
        ]
        self.ignored_elements = constants.TENTACLES_PACKAGE_IGNORED_ELEMENTS

    def should_ignore(self, folder_path, names):
        return [name
                for name in names
                if self._should_ignore(folder_path, name)]

    def _should_ignore(self, element_path, element_name):
        if element_name in self.ignored_elements:
            return True
        if self.tentacles_white_list is not None:
            candidate_path = path.join(element_path, element_name)
            return path.isdir(candidate_path) and candidate_path in self.tentacle_paths_black_list
        return False



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import os

import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.creators as tentacle_creator
import octobot_tentacles_manager.exporters as exporters
import octobot_tentacles_manager.models as models
import octobot_tentacles_manager.util as util
import octobot_tentacles_manager.api.uploader as uploader_api
import octobot_commons.logging as logging


def start_tentacle_creator(config, commands) -> int:
    tentacle_creator_inst = tentacle_creator.TentacleCreator(config)
    return tentacle_creator_inst.parse_commands(commands)


async def create_tentacles_package(package_name: str,
                                   tentacles_folder: str = constants.TENTACLES_PATH,
                                   output_dir: str = constants.DEFAULT_EXPORT_DIR,
                                   exported_tentacles_package: str = None,
                                   in_zip: bool = True,
                                   with_dev_mode: bool = False,
                                   upload_details: list = None,
                                   cythonize: bool = False) -> int:
    tentacle_package: models.TentaclePackage = models.TentaclePackage(package_name)
    export_result: int = await exporters.TentaclePackageExporter(artifact=tentacle_package,
                                                                 tentacles_folder=tentacles_folder,
                                                                 exported_tentacles_package=exported_tentacles_package,
                                                                 output_dir=output_dir,
                                                                 should_zip=in_zip,
                                                                 with_dev_mode=with_dev_mode,
                                                                 should_cythonize=cythonize).export()
    if upload_details is not None and len(upload_details) > 0:
        export_path: str = tentacle_package.output_path
        alias_name: str = upload_details[1] if len(upload_details) > 1 else os.path.basename(export_path)
        await uploader_api.upload_file_or_folder_to_nexus(nexus_path=upload_details[0],
                                                          artifact_path=export_path,
                                                          artifact_alias=alias_name)
    return export_result


async def create_all_tentacles_bundle(output_dir: str = constants.DEFAULT_EXPORT_DIR,
                                      tentacles_folder: str = constants.TENTACLES_PATH,
                                      exported_tentacles_package: str = None,
                                      in_zip: bool = True,
                                      with_dev_mode: bool = False,
                                      cythonize: bool = False,
                                      should_remove_artifacts_after_use: bool = False,
                                      upload_url: str = None,
                                      should_zip_bundle: bool = False) -> int:
    logger = logging.get_logger("TentacleManagerApi")
    error_count: int = 0
    tentacle_bundle_exported_list = []
    tentacles: list = util.load_tentacle_with_metadata(tentacles_folder)
    tentacles_white_list = util.filter_tentacles_by_dev_mode_and_package(
        tentacles=tentacles,
        with_dev_mode=with_dev_mode,
        package_filter=exported_tentacles_package
    )
    for tentacle in tentacles_white_list:
        try:
            tentacle_package = models.TentaclePackage()
            tentacle_exporter = exporters.TentacleExporter(artifact=tentacle,
                                                           output_dir=output_dir,
                                                           tentacles_folder=tentacles_folder,
                                                           should_zip=in_zip,
                                                           with_dev_mode=with_dev_mode,
                                                           should_cythonize=cythonize)
            await tentacle_exporter.export()
            tentacle_package.add_artifact(tentacle)
            tentacle_bundle_exporter = exporters.TentacleBundleExporter(
                artifact=tentacle_package,
                tentacles_folder=tentacles_folder,
                output_dir=output_dir,
                should_zip=should_zip_bundle,
                should_remove_artifacts_after_use=should_remove_artifacts_after_use)
            await tentacle_bundle_exporter.export()
            tentacle_bundle_exported_list.append(tentacle_bundle_exporter)
        except Exception as e:
            logger.error(f"Error when exporting tentacle {tentacle.name} : {str(e)}")
            error_count += 1

    if upload_url is not None:
        await asyncio.gather(
            *[_upload_exported_tentacle_bundle(upload_url, exported_tentacle_bundle)
              for exported_tentacle_bundle in tentacle_bundle_exported_list])

    return error_count


async def _upload_exported_tentacle_bundle(upload_url: str, exported_tentacle_bundle: exporters.TentacleBundleExporter):
    export_path = exported_tentacle_bundle.artifact.output_path
    alias_path = os.path.basename(export_path)
    if models.TentaclePackage.ARTIFACT_VERSION_SEPARATOR in export_path:
        alias_name, alias_version = alias_path.split(models.TentaclePackage.ARTIFACT_VERSION_SEPARATOR)
        alias_path = f"{alias_name}/{alias_version}"
    await uploader_api.upload_file_or_folder_to_nexus(nexus_path=upload_url,
                                                      artifact_path=export_path,
                                                      artifact_alias=alias_path)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_tentacles_manager.workers import tentacles_worker
from octobot_tentacles_manager.workers.tentacles_worker import (
    TentaclesWorker,
)

from octobot_tentacles_manager.workers import install_worker
from octobot_tentacles_manager.workers.install_worker import (
    InstallWorker,
)

from octobot_tentacles_manager.workers import uninstall_worker
from octobot_tentacles_manager.workers import repair_worker
from octobot_tentacles_manager.workers import update_worker
from octobot_tentacles_manager.workers import single_install_worker

from octobot_tentacles_manager.workers.uninstall_worker import (
    UninstallWorker,
)
from octobot_tentacles_manager.workers.repair_worker import (
    RepairWorker,
)
from octobot_tentacles_manager.workers.update_worker import (
    UpdateWorker,
)
from octobot_tentacles_manager.workers.single_install_worker import (
    SingleInstallWorker,
)

__all__ = [
    "UninstallWorker",
    "RepairWorker",
    "TentaclesWorker",
    "UpdateWorker",
    "InstallWorker",
    "SingleInstallWorker",
]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import os.path as path
import shutil

import octobot_commons.logging as logging
import octobot_commons.constants as commons_constants

import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.configuration as configuration
import octobot_tentacles_manager.loaders as loaders


class TentaclesSetupConfiguration:
    TENTACLE_ACTIVATION_KEY = "tentacle_activation"
    REGISTERED_TENTACLES_KEY = "registered_tentacles"
    DEFAULT_DEACTIVATABLE_TENTACLE_SUB_TYPES = {
        constants.TENTACLES_EVALUATOR_REALTIME_PATH,
        constants.TENTACLES_EVALUATOR_TA_PATH,
        constants.TENTACLES_EVALUATOR_SOCIAL_PATH,
        constants.TENTACLES_EVALUATOR_STRATEGIES_PATH,
        constants.TENTACLES_TRADING_MODE_PATH
    }

    def __init__(self, bot_installation_path=constants.DEFAULT_BOT_PATH,
                 config_path=constants.USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.bot_installation_path = bot_installation_path
        self.config_path = path.join(bot_installation_path, config_path)
        self.tentacles_activation = {}
        self.registered_tentacles = {}

    def get_config_folder(self) -> str:
        return path.split(self.config_path)[0]

    def register_tentacles_package(self, package_name, package_location):
        self.registered_tentacles[package_name] = package_location

    def unregister_tentacles_package(self, package_name):
        self.registered_tentacles.pop(package_name)

    async def fill_tentacle_config(self, tentacles, default_tentacle_config=constants.DEFAULT_TENTACLE_CONFIG,
                                   remove_missing_tentacles=True, update_location=None,
                                   force_update_registered_tentacles=False, newly_installed_tentacles=None,
                                   uninstalled_tentacles=None):
        self._update_tentacles_setup_config(tentacles,
                                            default_tentacle_config_file=default_tentacle_config,
                                            remove_missing_tentacles=remove_missing_tentacles,
                                            newly_installed_tentacles=newly_installed_tentacles,
                                            uninstalled_tentacles=uninstalled_tentacles)
        if update_location or force_update_registered_tentacles:
            self._update_registered_tentacles(tentacles, update_location)

    def refresh_profile_tentacles_config(self,
                                         tentacles,
                                         profiles_path=commons_constants.USER_PROFILES_FOLDER,
                                         newly_installed_tentacles=None,
                                         uninstalled_tentacles=None
                                         ):
        bot_profiles_path = os.path.join(self.bot_installation_path, profiles_path)
        if path.isdir(bot_profiles_path):
            for profile_folder in os.scandir(bot_profiles_path):
                self._refresh_profile_tentacles_config(tentacles, profile_folder,
                                                       newly_installed_tentacles, uninstalled_tentacles)

    def update_activation_configuration(self, new_config, deactivate_other_evaluators,
                                        add_missing_elements, tentacles_path=constants.TENTACLES_PATH):
        something_changed = False
        loaders.ensure_tentacles_metadata(tentacles_path)
        for element_name, activated in new_config.items():
            try:
                element_type = loaders.get_tentacle_classes()[element_name].tentacle_root_type
                if element_name in self.tentacles_activation[element_type]:
                    current_activation = self.tentacles_activation[element_type][element_name]
                    if current_activation != activated:
                        self.logger.info(f"Tentacles configuration updated: {element_name} "
                                         f"{'activated' if activated else 'deactivated'}")
                        self.tentacles_activation[element_type][element_name] = activated
                        something_changed = True
                elif add_missing_elements:
                    self.tentacles_activation[element_type][element_name] = activated
                    something_changed = True
            except KeyError:
                # tentacle missing in loaded metadata: can't be used
                self.logger.error(f"Impossible to change activation of {element_name}: this Tentacle class is not "
                                  f"in loaded tentacles metadata dict.")
        if deactivate_other_evaluators:
            something_changed = self._deactivate_other_evaluators(new_config) or something_changed
        return something_changed

    def is_tentacle_activated(self, tentacle_class_name, tentacles_path=constants.TENTACLES_PATH):
        loaders.ensure_tentacles_metadata(tentacles_path)
        tentacle_type = loaders.get_tentacle_classes()[tentacle_class_name].tentacle_root_type
        return self.tentacles_activation[tentacle_type][tentacle_class_name]

    def from_activated_tentacles_classes(self, *tentacles_classes, tentacles_path=constants.TENTACLES_PATH):
        loaders.ensure_tentacles_metadata(tentacles_path)
        tentacle_by_class_name = loaders.get_tentacle_classes()
        for tentacle_class in tentacles_classes:
            tentacle_name = tentacle_class.get_name()
            tentacle = tentacle_by_class_name[tentacle_name]
            try:
                self.tentacles_activation[tentacle.tentacle_root_type][tentacle_name] = True
            except KeyError:
                self.tentacles_activation[tentacle.tentacle_root_type] = {tentacle_name: True}

    def _deactivate_other_evaluators(self, new_config):
        something_changed = False
        for element_type, element_names in self.tentacles_activation.items():
            for element_name in element_names:
                # for each activated element that is not in new_config
                if element_name not in new_config and self.tentacles_activation[element_type][element_name]:
                    # only handle evaluator tentacles
                    something_changed = self._deactivate_tentacle_if_evaluator(element_name, element_type) or \
                                        something_changed

        return something_changed

    def _deactivate_tentacle_if_evaluator(self, element_name, element_type):
        try:
            if loaders.get_tentacle_classes()[element_name].get_simple_tentacle_type() in {
                constants.TENTACLES_EVALUATOR_TA_PATH,
                constants.TENTACLES_EVALUATOR_SOCIAL_PATH,
                constants.TENTACLES_EVALUATOR_REALTIME_PATH
            }:
                self.logger.info(f"Tentacles configuration updated: {element_name} {'deactivated'}")
                self.tentacles_activation[element_type][element_name] = False
                return True
        except KeyError:
            # might happen when a tentacle is listed in config but has not been loaded in OctoBot (import fail etc)
            pass
        return False

    def read_config(self, tentacles_path=constants.TENTACLES_PATH):
        try:
            self._from_dict(configuration.read_config(self.config_path))
        except Exception as e:
            self.logger.error(f"Error when reading tentacles global configuration file ({e}), "
                              "resetting this file with default values. This will not change "
                              "any specific tentacle configuration.")
            loaders.ensure_tentacles_metadata(tentacles_path)
            self._update_tentacles_setup_config(loaders.get_tentacle_classes().values())
            self.save_config()

    def save_config(self):
        parent_dir, _ = path.split(self.config_path)
        if not path.exists(parent_dir):
            os.makedirs(parent_dir)
        configuration.write_config(self.config_path, self._to_dict())

    def _update_tentacles_setup_config(self,
                                       tentacles,
                                       default_tentacle_config_file=constants.DEFAULT_TENTACLE_CONFIG,
                                       remove_missing_tentacles=True,
                                       newly_installed_tentacles=None,
                                       uninstalled_tentacles=None):
        default_config = configuration.read_config(default_tentacle_config_file)
        default_activation_config = default_config[self.TENTACLE_ACTIVATION_KEY] \
            if self.TENTACLE_ACTIVATION_KEY in default_config else {}
        for tentacle in tentacles:
            self._update_tentacle_activation(tentacle, default_activation_config)
        if remove_missing_tentacles:
            self._filter_tentacle_activation(tentacles)
        if newly_installed_tentacles or uninstalled_tentacles:
            self._update_tentacles_groups_activation(tentacles,
                                                     newly_installed_tentacles,
                                                     uninstalled_tentacles)

    def _update_tentacle_activation(self, tentacle, default_config):
        if tentacle.tentacle_root_type not in self.tentacles_activation:
            self.tentacles_activation[tentacle.tentacle_root_type] = {}
        for tentacle_class_name in tentacle.tentacle_class_names:
            if tentacle_class_name not in self.tentacles_activation[tentacle.tentacle_root_type]:
                self._set_activation_using_default_config(tentacle, tentacle_class_name, default_config)

    def _set_activation_using_default_config(self, tentacle, tentacle_class_name, default_config):
        if tentacle.tentacle_root_type in default_config:
            # if tentacle_type in default config: use default value or do not activate (unless
            # if tentacle sub type in DEFAULT_ACTIVATED_TENTACLE_TYPES)
            if tentacle_class_name in default_config[tentacle.tentacle_root_type]:
                self.tentacles_activation[tentacle.tentacle_root_type][tentacle_class_name] = \
                    default_config[tentacle.tentacle_root_type][tentacle_class_name]
            else:
                # activate by default unless the sub type of this tentacle in among the sub types to be deactivated by
                # default
                self.tentacles_activation[tentacle.tentacle_root_type][tentacle_class_name] = \
                    tentacle.get_simple_tentacle_type() not in self.DEFAULT_DEACTIVATABLE_TENTACLE_SUB_TYPES
        else:
            # if tentacle_type not in default config: activate by default
            self.tentacles_activation[tentacle.tentacle_root_type][tentacle_class_name] = True

    def _filter_tentacle_activation(self, tentacles):
        tentacle_names = [tentacle_class_name
                          for tentacle in tentacles
                          for tentacle_class_name in tentacle.tentacle_class_names]
        for element_type, element_names in self.tentacles_activation.items():
            for element_name in list(element_names):
                if element_name not in tentacle_names:
                    self.tentacles_activation[element_type].pop(element_name)

    def _update_tentacles_groups_activation(self, tentacles, newly_installed_tentacles, uninstalled_tentacles):
        if newly_installed_tentacles:
            for new_tentacle in newly_installed_tentacles:
                if new_tentacle.get_simple_tentacle_type() not in self.DEFAULT_DEACTIVATABLE_TENTACLE_SUB_TYPES \
                        and new_tentacle.tentacle_group != new_tentacle.name:
                    # activate new_tentacle if part of tentacle to activate by default and has a tentacle group from
                    # which it's not the default tentacle: to avoid double tentacles: deactivate default one
                    self._update_default_tentacles_activation_for_group(tentacles, new_tentacle, False)
        if uninstalled_tentacles:
            for removed_tentacle in uninstalled_tentacles:
                if removed_tentacle.get_simple_tentacle_type() not in self.DEFAULT_DEACTIVATABLE_TENTACLE_SUB_TYPES \
                        and removed_tentacle.tentacle_group != removed_tentacle.name:
                    # re-activate default tentacle
                    self._update_default_tentacles_activation_for_group(tentacles, removed_tentacle, True)

    def _update_default_tentacles_activation_for_group(self, tentacles, updated_tentacle, activate):
        for tentacle in tentacles:
            if tentacle.tentacle_group == updated_tentacle.tentacle_group and tentacle.name != updated_tentacle.name:
                for tentacle_class_name in tentacle.tentacle_class_names:
                    self.tentacles_activation[tentacle.tentacle_root_type][tentacle_class_name] = activate

    def _update_registered_tentacles(self, tentacles, update_location):
        packages = set(tentacle.origin_package
                       for tentacle in tentacles)
        for package in packages:
            if package not in self.registered_tentacles:
                self.register_tentacles_package(package,
                                                update_location or constants.UNKNOWN_TENTACLES_PACKAGE_LOCATION)
        for registered_package in list(self.registered_tentacles):
            if registered_package not in packages:
                self.unregister_tentacles_package(registered_package)

    # Profiles management
    def _refresh_profile_tentacles_config(self, tentacles, profile_folder,
                                          newly_installed_tentacles, uninstalled_tentacles):
        self._refresh_profile_tentacles_config_file(tentacles,
                                                    path.join(profile_folder,
                                                              commons_constants.CONFIG_TENTACLES_FILE),
                                                    newly_installed_tentacles,
                                                    uninstalled_tentacles)
        self._refresh_profile_tentacles_specific_config(path.join(profile_folder,
                                                                  constants.TENTACLES_SPECIFIC_CONFIG_FOLDER))

    def _refresh_profile_tentacles_config_file(self, tentacles, tentacles_config_file,
                                               newly_installed_tentacles, uninstalled_tentacles):
        profile_setup_config = TentaclesSetupConfiguration(config_path=tentacles_config_file)
        profile_setup_config._from_dict(configuration.read_config(profile_setup_config.config_path))
        # use self.registered_tentacles as a reference
        profile_setup_config.registered_tentacles = self.registered_tentacles
        self._refresh_other_tentacles_activation_from_self(profile_setup_config)
        profile_setup_config._update_tentacles_groups_activation(tentacles,
                                                                 newly_installed_tentacles,
                                                                 uninstalled_tentacles)
        profile_setup_config.save_config()

    def _refresh_other_tentacles_activation_from_self(self, other):
        # add any missing element in tentacles activation
        for element_type, element_names in self.tentacles_activation.items():
            if element_type not in other.tentacles_activation:
                other.tentacles_activation[element_type] = {}
            for element_name in element_names:
                if element_name not in other.tentacles_activation[element_type]:
                    other.tentacles_activation[element_type][element_name] = \
                        self.tentacles_activation[element_type][element_name]
        # remove deleted tentacles activation
        for element_type, element_names in other.tentacles_activation.items():
            to_remove = []
            for element_name in element_names:
                try:
                    self.tentacles_activation[element_type][element_name]
                except KeyError:
                    to_remove.append(element_name)
            for element in to_remove:
                element_names.pop(element, None)

    def _refresh_profile_tentacles_specific_config(self, tentacles_specific_config_folder):
        tentacles_config_path = path.join(path.split(self.config_path)[0],
                                          constants.TENTACLES_SPECIFIC_CONFIG_FOLDER)
        if not path.isdir(tentacles_specific_config_folder):
            os.mkdir(tentacles_specific_config_folder)
        profile_tentacles_configs = os.listdir(tentacles_specific_config_folder)
        for tentacle_config in os.scandir(tentacles_config_path):
            # add missing specific tentacles config
            if tentacle_config.name not in profile_tentacles_configs:
                shutil.copy(tentacle_config, path.join(tentacles_specific_config_folder, tentacle_config.name))

    def _from_dict(self, input_dict):
        if self.TENTACLE_ACTIVATION_KEY in input_dict:
            self.tentacles_activation = input_dict[self.TENTACLE_ACTIVATION_KEY]
        if self.REGISTERED_TENTACLES_KEY in input_dict:
            self.registered_tentacles = input_dict[self.REGISTERED_TENTACLES_KEY]

    def _to_dict(self):
        return {
            self.TENTACLE_ACTIVATION_KEY: self.tentacles_activation,
            self.REGISTERED_TENTACLES_KEY: self.registered_tentacles
        }



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import os.path as path

import octobot_tentacles_manager.managers as managers
import octobot_tentacles_manager.workers as workers
import octobot_tentacles_manager.util as util


class RepairWorker(workers.TentaclesWorker):

    def __init__(self,
                 reference_tentacles_dir,
                 tentacle_path,
                 bot_installation_path,
                 use_confirm_prompt,
                 aiohttp_session,
                 bot_install_dir):
        super().__init__(reference_tentacles_dir,
                         tentacle_path,
                         bot_installation_path,
                         use_confirm_prompt,
                         aiohttp_session,
                         bot_install_dir=bot_install_dir)
        self.verbose = True

    async def process(self, name_filter=None) -> int:
        # force reset of all init files
        await self.tentacles_setup_manager.remove_tentacle_arch_init_files()
        # create missing files and folders
        await self.tentacles_setup_manager.create_missing_tentacles_arch()
        self.reset_worker()
        self.progress = 1
        self.available_tentacles = util.load_tentacle_with_metadata(self.tentacle_path)
        self.total_steps = len(self.available_tentacles)
        await asyncio.gather(*[self._repair_tentacle(tentacle) for tentacle in self.available_tentacles])
        await self.tentacles_setup_manager.refresh_user_tentacles_setup_config_file(
            force_update_registered_tentacles=True
        )
        self.log_summary()
        return len(self.errors)

    async def _repair_tentacle(self, tentacle):
        try:
            tentacle_module_path = path.join(tentacle.tentacle_path, tentacle.name)
            tentacle_manager = managers.TentacleManager(tentacle, self.bot_installation_path)
            await managers.create_tentacle_init_file_if_necessary(tentacle_module_path, tentacle)
            tentacle_manager.import_tentacle_config_if_any(tentacle_module_path)
            managers.update_tentacle_type_init_file(tentacle, tentacle.tentacle_path)
            if self.verbose:
                self.logger.info(f"[{self.progress}/{self.total_steps}] {tentacle} ready to use")
        except Exception as e:
            message = f"Error when repairing {tentacle.name}: {e}"
            self.errors.append(message)
            self.logger.exception(e, True, message)
        finally:
            self.progress += 1



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import distutils.version as loose_version

import octobot_commons.logging as logging

import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.loaders as loaders


def get_installed_tentacles_modules() -> set:
    return set(tentacle for tentacle in loaders.get_tentacle_classes().values())


def get_tentacle_group(klass) -> str:
    return loaders.get_tentacle(klass).tentacle_group


def get_tentacle_version(klass) -> str:
    return loaders.get_tentacle(klass).version


def get_tentacle_origin_package(klass) -> str:
    return loaders.get_tentacle(klass).origin_package


def get_tentacle_module_name(klass) -> str:
    return loaders.get_tentacle(klass).name


def get_tentacle_resources_path(klass) -> str:
    return loaders.get_resources_path(klass)


def get_tentacle_documentation_path(klass) -> str:
    return loaders.get_documentation_file_path(klass)


def check_tentacle_version(version, name, origin_package, verbose=True) -> bool:
    logger = logging.get_logger("TentacleChecker")
    try:
        if origin_package == constants.DEFAULT_TENTACLES_PACKAGE:
            if loose_version.LooseVersion(version) < loose_version.LooseVersion(
                    constants.TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION) and verbose:
                logger.error(f"Incompatible tentacle {name}: version {version}, "
                             f"minimum expected: {constants.TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION} "
                             f"this tentacle may not work properly. "
                             f"Please update your tentacles ('start.py -p update {name}' "
                             f"or 'start.py -p update all')")
                return False
    except Exception as e:
        if verbose:
            logger.error(f"Error when reading tentacle metadata: {e}")
    return True



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os.path as path
import shutil

import octobot_tentacles_manager.configuration as configuration
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.loaders as loaders


def get_config(tentacles_setup_config, klass) -> dict:
    return configuration.read_config(_get_config_file_path(tentacles_setup_config, klass))


def update_config(tentacles_setup_config, klass, config_update) -> None:
    config_file = _get_config_file_path(tentacles_setup_config, klass)
    current_config = configuration.read_config(config_file)
    # only update values in config update not to erase values in root config (might not be editable)
    current_config.update(config_update)
    configuration.write_config(config_file, current_config)


def factory_reset_config(tentacles_setup_config, klass) -> None:
    reference_config_path = path.join(_get_reference_config_path(klass), _get_config_file_name(klass))
    shutil.copy(reference_config_path, _get_config_file_path(tentacles_setup_config, klass))


def get_config_schema_path(klass) -> str:
    return path.join(_get_reference_config_path(klass), f"{klass.get_name()}{constants.CONFIG_SCHEMA_EXT}")


def _get_config_file_path(tentacles_setup_config, klass) -> str:
    return path.join(tentacles_setup_config.get_config_folder(),
                     constants.TENTACLES_SPECIFIC_CONFIG_FOLDER,
                     _get_config_file_name(klass))


def _get_reference_config_path(klass) -> str:
    return path.join(loaders.get_tentacle_module_path(klass), constants.TENTACLE_CONFIG)


def _get_config_file_name(klass) -> str:
    return f"{klass.get_name()}{constants.CONFIG_EXT}"



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import abc
import os
import os.path as path
import shutil

import octobot_commons.logging as logging
import octobot_tentacles_manager.models as models
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.creators as creators
import octobot_tentacles_manager.managers as managers
import octobot_tentacles_manager.util as util


class ArtifactExporter:
    __metaclass__ = abc.ABCMeta

    def __init__(self,
                 artifact: models.Artifact,
                 tentacles_folder: str,
                 output_dir: str = constants.DEFAULT_EXPORT_DIR,
                 should_cythonize: bool = False,
                 should_zip: bool = False,
                 with_dev_mode: bool = False):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.artifact = artifact
        self.tentacles_folder: str = tentacles_folder
        self.output_dir: str = output_dir if output_dir is not None else constants.DEFAULT_EXPORT_DIR

        self.should_cleanup_working_folder: bool = False
        self.should_cythonize = should_cythonize
        self.should_zip = should_zip
        self.with_dev_mode = with_dev_mode

        self.artifact.output_dir = self.output_dir
        self.artifact.output_path = self.artifact.name
        self.working_folder = constants.TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER \
            if self.should_zip else os.path.join(self.output_dir, self.artifact.name)

    @abc.abstractmethod
    async def prepare_export(self):
        """
        Called before export process, should be implemented if necessary
        :return: None
        """
        raise NotImplementedError("prepare_export is not implemented")

    @abc.abstractmethod
    async def after_export(self) -> None:
        """
        Called after export process, should be implemented if necessary
        :return: None
        """
        raise NotImplementedError("after_export is not implemented")

    async def export(self) -> int:
        """
        Run artifact export process by calling :
        - export preparation with abstract prepare_export()
        - export cleanup, cythonization and compression with cleanup_cythonize_and_archive_if_necessary()
        - export finalization with abstract after_export()
        :return: an error code as integer (0 for success and 1 for fail)
        """
        try:
            await self.prepare_export()
            await self.cleanup_cythonize_and_archive_if_necessary()
            await self.after_export()
            return 0
        except Exception as e:
            self.logger.exception(e, True, f"Error while performing {self.artifact.ARTIFACT_NAME} export : {e}")
            return 1

    async def cleanup_cythonize_and_archive_if_necessary(self) -> None:
        """
        Perform cleanup, cythonization and compression if necessary
        :return: None
        """
        if self.should_cleanup_working_folder:
            await self.cleanup_working_folder()

        # handle tentacles cythonization if required
        if self.should_cythonize:
            await creators.cythonize_and_compile_tentacles(self.working_folder)

        # Zip if required
        if self.should_zip:
            self.artifact.output_path = self.zip_working_folder()
            self.logger.info(f"Zipped {self.artifact.ARTIFACT_NAME} available at: {self.artifact.output_path}")
        else:
            self.artifact.output_path = os.path.abspath(self.working_folder)
            self.logger.info(f"Cleaned {self.artifact.ARTIFACT_NAME} available at: {self.artifact.output_path}")

    async def cleanup_working_folder(self) -> None:
        """
        Clean working folder from tentacle file architecture and non tentacle files
        :return: None
        """
        # cleanup temp working folder
        tentacles_setup_manager = managers.TentaclesSetupManager(self.working_folder)
        await tentacles_setup_manager.remove_tentacle_arch_init_files()
        util.remove_unnecessary_files(self.working_folder)

        # remove non tentacles files before zipping
        if self.should_zip:
            util.remove_non_tentacles_files(self.working_folder, self.logger)

    @staticmethod
    def create_or_replace_temporary_creator_dir() -> None:
        """
        Remove temporary creator directory if exists and create it
        :return:
        """
        if path.exists(constants.TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER):
            shutil.rmtree(constants.TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)

    def copy_directory_content_to_temporary_dir(self, folder_to_copy: str, ignore=None) -> None:
        """
        Copy a directory content to creator temporary dir
        :param folder_to_copy: the folder to copy content
        :param ignore: shutil ignoring file pattern
        :return: None
        """
        self.create_or_replace_temporary_creator_dir()
        shutil.copytree(folder_to_copy, self.working_folder, ignore=ignore)

    def copy_directory_content_to_working_dir(self, folder_to_copy: str, ignore=None) -> None:
        """
        Copy a directory content to working dir
        :param folder_to_copy: the folder to copy content
        :param ignore: shutil ignoring file pattern
        :return: None
        """
        if not path.exists(self.working_folder):
            shutil.copytree(folder_to_copy, self.working_folder, ignore=ignore)
        else:
            util.merge_folders(folder_to_copy, self.working_folder, ignore)

    def zip_working_folder(self) -> str:
        """
        Archive creator temporary dir to a zip file
        :return: the path of the archive
        """
        # remove .zip extension if necessary
        file_name = self.artifact.name.split(f".{constants.TENTACLES_PACKAGE_FORMAT}")[0].\
            replace(models.TentaclePackage.ARTIFACT_VERSION_SEPARATOR, "_")
        zipped_file = shutil.make_archive(os.path.join(self.artifact.output_dir, file_name),
                                          constants.TENTACLES_PACKAGE_FORMAT, self.working_folder)
        try:
            # remove working folder
            shutil.rmtree(constants.TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)
        except Exception as e:
            self.logger.error(f"Error when cleaning up temporary folder: {e}")
        return zipped_file



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import os
import os.path as path
import shutil

import octobot_commons.logging as logging

import octobot_tentacles_manager.configuration as configuration
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.managers as managers
import octobot_tentacles_manager.util as util


class TentaclesSetupManager:

    def __init__(self, tentacle_setup_root_path,
                 bot_installation_path=constants.DEFAULT_BOT_PATH,
                 default_tentacle_config=constants.DEFAULT_TENTACLE_CONFIG):
        self.tentacle_setup_root_path = tentacle_setup_root_path
        self.default_tentacle_config = default_tentacle_config
        self.bot_installation_path = bot_installation_path

    async def refresh_user_tentacles_setup_config_file(self,
                                                       tentacles_setup_config_to_update=None,
                                                       update_location=None,
                                                       force_update_registered_tentacles=False,
                                                       newly_installed_tentacles=None,
                                                       uninstalled_tentacles=None):
        available_tentacle = util.load_tentacle_with_metadata(self.tentacle_setup_root_path)
        if not tentacles_setup_config_to_update:
            tentacle_setup_config = configuration.TentaclesSetupConfiguration(
                bot_installation_path=self.bot_installation_path)
            tentacle_setup_config.read_config(self.tentacle_setup_root_path)
        else:
            tentacle_setup_config = tentacles_setup_config_to_update
        await tentacle_setup_config.fill_tentacle_config(
            available_tentacle,
            self.default_tentacle_config,
            update_location=update_location,
            force_update_registered_tentacles=force_update_registered_tentacles,
            newly_installed_tentacles=newly_installed_tentacles,
            uninstalled_tentacles=uninstalled_tentacles
        )
        tentacle_setup_config.save_config()
        tentacle_setup_config.refresh_profile_tentacles_config(
            available_tentacle,
            newly_installed_tentacles=newly_installed_tentacles,
            uninstalled_tentacles=uninstalled_tentacles
        )

    async def create_missing_tentacles_arch(self):
        # tentacle user config folder
        await util.find_or_create(path.join(self.bot_installation_path, constants.USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH))
        # tentacles folder
        found_existing_installation = not await util.find_or_create(self.tentacle_setup_root_path)
        # tentacle main python init file
        await util.find_or_create(path.join(self.tentacle_setup_root_path, constants.PYTHON_INIT_FILE), False,
                                  managers.get_module_init_file_content(constants.TENTACLES_FOLDERS_ARCH.keys()))
        # tentacle inner architecture
        await TentaclesSetupManager._tentacle_arch_operation(self.tentacle_setup_root_path,
                                                             constants.TENTACLES_FOLDERS_ARCH,
                                                             self._create_missing_files_and_folders,
                                                             util.find_or_create)
        await self._create_missing_files_and_folders(self.tentacle_setup_root_path, constants.TENTACLES_FOLDERS_ARCH)
        return found_existing_installation

    async def remove_tentacle_arch_init_files(self):
        await TentaclesSetupManager._remove_tentacles_arch_init_file(self.tentacle_setup_root_path, None)
        await TentaclesSetupManager._tentacle_arch_operation(self.tentacle_setup_root_path,
                                                             constants.TENTACLES_FOLDERS_ARCH,
                                                             self._remove_tentacles_arch_init_file,
                                                             self._remove_tentacles_arch_init_file)

    @staticmethod
    def cleanup_temp_dirs():
        if path.exists(constants.TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR):
            shutil.rmtree(constants.TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR)

    @staticmethod
    def is_tentacles_arch_valid(verbose=True, raises=False) -> bool:
        try:
            if not TentaclesSetupManager._is_full_arch_valid(constants.TENTACLES_PATH,
                                                             constants.TENTACLES_FOLDERS_ARCH):
                return False
            import tentacles
            return True
        except (ImportError, SyntaxError) as e:
            if verbose:
                logging.get_logger(TentaclesSetupManager.__name__).exception(e, True,
                                                                             f"Error when importing tentacles: {e}")
            if raises:
                raise e
            return False

    @staticmethod
    def _is_full_arch_valid(root_folder, files_arch):
        for root, modules in files_arch.items():
            current_root = path.join(root_folder, root)
            if isinstance(modules, dict):
                if not TentaclesSetupManager._is_full_arch_valid(current_root, modules):
                    return False
            elif not all(path.exists(path.join(current_root, directory)) for directory in modules):
                return False
        return True

    @staticmethod
    def delete_tentacles_arch(force=False, raises=False,
                              with_user_config=False, bot_installation_path=constants.DEFAULT_BOT_PATH):
        if TentaclesSetupManager.is_tentacles_arch_valid(verbose=False, raises=raises) \
                or (force and path.exists(path.join(bot_installation_path, constants.TENTACLES_PATH))):
            shutil.rmtree(path.join(bot_installation_path, constants.TENTACLES_PATH))
        if with_user_config and path.exists(path.join(bot_installation_path, constants.USER_REFERENCE_TENTACLE_CONFIG_PATH)):
            shutil.rmtree(path.join(bot_installation_path, constants.USER_REFERENCE_TENTACLE_CONFIG_PATH))

    @staticmethod
    def get_available_tentacles_repos():
        # TODO: add advanced tentacles repos
        return [constants.TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR]

    @staticmethod
    async def _tentacle_arch_operation(root_folder, files_arch, branch_func, leaf_func):
        sub_dir_to_handle_coroutines = []
        for root, modules in files_arch.items():
            current_root = path.join(root_folder, root)
            await branch_func(current_root, modules)
            if isinstance(modules, dict):
                sub_dir_to_handle_coroutines.append(
                    TentaclesSetupManager._tentacle_arch_operation(current_root, modules, branch_func, leaf_func))
            elif leaf_func is not None:
                sub_dir_to_handle_coroutines += [leaf_func(path.join(current_root, directory))
                                                 for directory in modules]
        await asyncio.gather(*sub_dir_to_handle_coroutines)

    @staticmethod
    async def _create_missing_files_and_folders(current_root, modules):
        await util.find_or_create(current_root)
        # create python init file
        await managers.find_or_create_module_init_file(current_root, modules)

    @staticmethod
    async def _remove_tentacles_arch_init_file(current_root, _=None):
        potential_init_file = path.join(current_root, constants.PYTHON_INIT_FILE)
        if path.exists(potential_init_file):
            os.remove(potential_init_file)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
from os.path import exists
from shutil import rmtree
from zipfile import ZipFile

import aiohttp
import pytest
from os import path, walk

import octobot_commons.constants as commons_constants
from octobot_tentacles_manager.api.installer import install_all_tentacles, install_tentacles, install_single_tentacle, \
    repair_installation
from octobot_tentacles_manager.configuration.tentacles_setup_configuration import TentaclesSetupConfiguration
from octobot_tentacles_manager.constants import TENTACLES_PATH, TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR, \
    PYTHON_INIT_FILE, TENTACLES_NOTIFIERS_PATH, USER_REFERENCE_TENTACLE_CONFIG_PATH, \
    USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH, TENTACLES_SERVICES_PATH, TENTACLES_BACKTESTING_PATH, TENTACLES_EVALUATOR_PATH
from octobot_tentacles_manager.managers.tentacles_setup_manager import TentaclesSetupManager
from tests import event_loop


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_install_all_tentacles():
    _cleanup(False)
    async with aiohttp.ClientSession() as session:
        assert await install_all_tentacles(_tentacles_local_path(), aiohttp_session=session) == 0
    _cleanup()


async def test_install_one_tentacle_with_requirement():
    async with aiohttp.ClientSession() as session:
        assert await install_tentacles(["reddit_service_feed"], _tentacles_local_path(), aiohttp_session=session) == 0
    assert path.exists(path.join(TENTACLES_PATH, "Services", "Services_bases", "reddit_service", "reddit_service.py"))
    _cleanup()


async def test_install_single_tentacle():
    tentacle_path = path.join("tests", "static", "momentum_evaluator")
    tentacle_type = "Evaluator/TA"
    async with aiohttp.ClientSession() as session:
        assert await install_single_tentacle(tentacle_path, tentacle_type, aiohttp_session=session) == 0
    assert path.exists(path.join(TENTACLES_PATH, "Evaluator", "TA", "momentum_evaluator", "momentum_evaluator.py"))
    assert not path.exists(TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR)
    # check availability of tentacle arch, installed momentum_evaluator and its reddit_service fake requirement
    assert len(list(walk(TENTACLES_PATH))) == 32
    _cleanup()


async def test_repair_installation():
    broken_install = "broken_installation"
    broken_tentacles = path.join("broken_installation", TENTACLES_PATH)
    with ZipFile(path.join("tests", "static", "broken_tentacles.zip")) as zipped_broken_tentacles:
        zipped_broken_tentacles.extractall(broken_install)

    assert await repair_installation(bot_path=broken_install) == 0

    # restore Notifiers directory
    assert path.isdir(path.join(broken_tentacles, TENTACLES_SERVICES_PATH, TENTACLES_NOTIFIERS_PATH))

    # restore backtesting init file
    assert path.isfile(path.join(broken_tentacles, TENTACLES_BACKTESTING_PATH, PYTHON_INIT_FILE))

    # restore main __init__ content
    with open(path.join(broken_tentacles, PYTHON_INIT_FILE)) as f:
        stripped_lines = [line.strip() for line in f.readlines()]
        assert "from .Trading import *" in stripped_lines
        assert "from .Backtesting import *" in stripped_lines
        assert "from .Interfaces import *" not in stripped_lines

    # restore Evaluator/Realtime  __init__ content
    rt_path = path.join(broken_tentacles, "Evaluator", "RealTime")
    with open(path.join(rt_path, PYTHON_INIT_FILE)) as f:
        stripped_lines = [line.strip() for line in f.readlines()]
        assert "from .other_instant_fluctuations_evaluator import *" in stripped_lines
        # did not add existing import twice
        assert sum([1
                    for line in stripped_lines
                    if "from .instant_fluctuations_evaluator" in line]) == 1

    # restore Evaluator/Realtime/other_instant_fluctuations_evaluator  __init__ file and content
    with open(path.join(rt_path, "other_instant_fluctuations_evaluator", PYTHON_INIT_FILE)) as f:
        assert "from .other_instant_fluctuations_evaluator import " \
               "OtherInstantFluctuationsEvaluator, SecondOtherInstantFluctuationsEvaluator" in f.readlines()

    # restore tentacles_config.json validity and content
    user_config_path = path.join(broken_install, USER_REFERENCE_TENTACLE_CONFIG_PATH)
    with open(path.join(user_config_path, commons_constants.CONFIG_TENTACLES_FILE)) as f:
        activations = json.load(f)[TentaclesSetupConfiguration.TENTACLE_ACTIVATION_KEY]
        # SecondOtherInstantFluctuationsEvaluator is activated because there is no available default config in
        # this context
        assert activations[TENTACLES_EVALUATOR_PATH]["SecondOtherInstantFluctuationsEvaluator"] is True

    # restore DailyTradingMode config file
    assert path.isfile(path.join(broken_install, USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH, "DailyTradingMode.json"))

    rmtree(broken_install)


def _tentacles_local_path():
    return path.join("tests", "static", "tentacles.zip")


def _cleanup(raises=True):
    if exists(TENTACLES_PATH):
        TentaclesSetupManager.delete_tentacles_arch(force=True, raises=raises)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import aiohttp
import pytest
import os
from logging import INFO

import octobot_commons.constants as commons_constants
from octobot_commons.logging.logging_util import set_logging_level
from octobot_tentacles_manager.constants import USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH, \
    TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR, USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, TENTACLES_PATH, DEFAULT_BOT_PATH, \
    UNKNOWN_TENTACLES_PACKAGE_LOCATION, TENTACLES_SPECIFIC_CONFIG_FOLDER
from octobot_tentacles_manager.workers.install_worker import InstallWorker
from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.util.tentacle_fetching import fetch_and_extract_tentacles
from tests import event_loop, clean, fake_profiles, TEMP_DIR, OTHER_PROFILE

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_install_two_tentacles(clean):
    _enable_loggers()
    tentacles_path = os.path.join("tests", "static", "tentacles.zip")
    await fetch_and_extract_tentacles(TEMP_DIR, tentacles_path, None)
    worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_path_or_url = tentacles_path
    worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    assert await worker.process(["instant_fluctuations_evaluator", "generic_exchange_importer"]) == 0

    # test installed files
    trading_mode_files_count = sum(1 for _ in os.walk(os.path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 1
    backtesting_mode_files_count = sum(1 for _ in os.walk(os.path.join(TENTACLES_PATH, "Backtesting", "importers")))
    assert backtesting_mode_files_count == 7
    config_files = [f for f in os.walk(USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH)]
    config_files_count = len(config_files)
    assert config_files_count == 1
    assert "InstantFluctuationsEvaluator.json" in config_files[0][2]
    assert "DailyTradingMode.json" not in config_files[0][2]
    assert len(config_files[0][2]) == 1

    # test tentacles config
    with open(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        ref_profile_config = json.load(config_f)
        assert ref_profile_config == {
            'registered_tentacles': {
                'OctoBot-Default-Tentacles': tentacles_path
            },
            'tentacle_activation': {
                'Backtesting': {
                    'GenericExchangeDataImporter': True
                },
                'Evaluator': {
                    'InstantFluctuationsEvaluator': True
                }
            }
        }


async def test_install_one_tentacle_with_requirement(clean):
    async with aiohttp.ClientSession() as session:
        _enable_loggers()
        await fetch_and_extract_tentacles(TEMP_DIR, os.path.join("tests", "static", "tentacles.zip"), None)
        worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, session)
        worker.tentacles_setup_manager.default_tentacle_config = \
            os.path.join("tests", "static", "default_tentacle_config.json")
        assert await worker.process(["reddit_service_feed"]) == 0

    # test removed temporary requirements files
    assert not os.path.exists(TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR)

    # test installed files
    trading_mode_files_count = sum(1 for _ in os.walk(os.path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 1
    config_files = [f for f in os.walk(USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH)]
    assert len(config_files) == 1
    assert len(config_files[0][2]) == 0

    # test tentacles config
    with open(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'registered_tentacles': {
                'OctoBot-Default-Tentacles': UNKNOWN_TENTACLES_PACKAGE_LOCATION
            },
            'tentacle_activation': {
                'Services': {
                    'RedditService': True,
                    'RedditServiceFeed': True
                }
            }
        }
    assert os.path.exists(
        os.path.join("tentacles", "Services", "Services_bases", "reddit_service", "reddit_service.py"))


async def test_install_all_tentacles(clean):
    _enable_loggers()
    tentacles_path = os.path.join("tests", "static", "tentacles.zip")
    await fetch_and_extract_tentacles(TEMP_DIR, tentacles_path, None)
    worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_path_or_url = tentacles_path
    worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    assert await worker.process() == 0

    # test installed files
    trading_mode_files_count = sum(1 for _ in os.walk(os.path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 5
    config_files = [f for f in os.walk(USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH)]
    config_files_count = len(config_files)
    assert config_files_count == 1
    assert "DailyTradingMode.json" in config_files[0][2]
    assert len(config_files[0][2]) == 5

    # test tentacles config
    with open(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'registered_tentacles': {
                'OctoBot-Default-Tentacles': tentacles_path
            },
            'tentacle_activation': {
                'Backtesting': {
                    'GenericExchangeDataImporter': True
                },
                'Evaluator': {
                    'InstantFluctuationsEvaluator': True,
                    'OtherInstantFluctuationsEvaluator': False,
                    'OverallStateAnalyser': True,
                    'RedditForumEvaluator': False,
                    'SecondOtherInstantFluctuationsEvaluator': False,
                    'SimpleMixedStrategyEvaluator': True,
                    'TextAnalysis': True
                },
                'Services': {
                    'RedditService': True,
                    'RedditServiceFeed': True
                },
                'Trading': {
                    'DailyTradingMode': True
                }
            }
        }


async def test_install_all_tentacles_with_profile(clean):
    _enable_loggers()
    profile_path = os.path.join(commons_constants.USER_PROFILES_FOLDER, "many_traded_elements")
    assert not os.path.isfile(os.path.join(profile_path, commons_constants.PROFILE_CONFIG_FILE))
    tentacles_path = os.path.join("tests", "static", "tentacles_with_profile.zip")
    await fetch_and_extract_tentacles(TEMP_DIR, tentacles_path, None)
    worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_path_or_url = tentacles_path
    worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    assert await worker.process() == 0

    # test installed files to ensure tentacles installation got well
    trading_mode_files_count = sum(1 for _ in os.walk(os.path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 5
    config_files = [f for f in os.walk(USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH)]
    config_files_count = len(config_files)
    assert config_files_count == 1
    assert "DailyTradingMode.json" in config_files[0][2]
    assert len(config_files[0][2]) == 5

    # test installed profile
    assert os.path.isfile(os.path.join(profile_path, commons_constants.PROFILE_CONFIG_FILE))
    assert os.path.isfile(os.path.join(profile_path, "default_profile.png"))
    assert os.path.isfile(os.path.join(profile_path, commons_constants.CONFIG_TENTACLES_FILE))
    assert os.path.isfile(os.path.join(profile_path, TENTACLES_SPECIFIC_CONFIG_FOLDER, "DailyTradingMode.json"))
    assert os.path.isfile(os.path.join(profile_path, TENTACLES_SPECIFIC_CONFIG_FOLDER, "TwitterNewsEvaluator.json"))


async def test_profiles_update(clean, fake_profiles):
    _enable_loggers()
    tentacles_path = os.path.join("tests", "static", "tentacles.zip")
    await fetch_and_extract_tentacles(TEMP_DIR, tentacles_path, None)
    worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_path_or_url = tentacles_path
    worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    # install all tentacles
    assert await worker.process() == 0

    # test tentacles setup config
    with open(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH) as config_f:
        ref_profile_config = json.load(config_f)

        # test profiles tentacles config
        with open(os.path.join(commons_constants.USER_PROFILES_FOLDER,
                               commons_constants.DEFAULT_PROFILE,
                               commons_constants.CONFIG_TENTACLES_FILE)) as default_c:
            assert ref_profile_config == json.load(default_c)
        with open(os.path.join(commons_constants.USER_PROFILES_FOLDER,
                               OTHER_PROFILE,
                               commons_constants.CONFIG_TENTACLES_FILE)) as other_c:
            assert ref_profile_config == json.load(other_c)

    # test specific tentacles config
    default_profile_tentacles_config = os.path.join(commons_constants.USER_PROFILES_FOLDER,
                                                    commons_constants.DEFAULT_PROFILE,
                                                    TENTACLES_SPECIFIC_CONFIG_FOLDER)
    other_profile_tentacles_config = os.path.join(commons_constants.USER_PROFILES_FOLDER,
                                                  OTHER_PROFILE,
                                                  TENTACLES_SPECIFIC_CONFIG_FOLDER)
    for tentacle_config in os.scandir(os.path.join(os.path.split(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH)[0],
                                                   TENTACLES_SPECIFIC_CONFIG_FOLDER)):
        with open(tentacle_config) as ref_config_file:
            ref_config = json.load(ref_config_file)
        with open(os.path.join(default_profile_tentacles_config, tentacle_config.name)) as default_profile_config_file:
            assert ref_config == json.load(default_profile_config_file)
        with open(os.path.join(other_profile_tentacles_config, tentacle_config.name)) as other_profile_config_file:
            assert ref_config == json.load(other_profile_config_file)


async def test_install_all_tentacles_twice(clean):
    await fetch_and_extract_tentacles(TEMP_DIR, os.path.join("tests", "static", "tentacles.zip"), None)
    worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    assert await worker.process() == 0
    assert await worker.process() == 0
    trading_mode_files_count = sum(1 for _ in os.walk(os.path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 5


async def test_install_all_tentacles_fetching_requirements(clean):
    async with aiohttp.ClientSession() as session:
        _enable_loggers()
        await fetch_and_extract_tentacles(TEMP_DIR, os.path.join("tests", "static", "requirements_tentacles.zip"), None)
        worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, session)
        worker.tentacles_setup_manager.default_tentacle_config = \
            os.path.join("tests", "static", "default_tentacle_config.json")
        assert await worker.process() == 0

    trading_mode_files_count = sum(1 for _ in os.walk(os.path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 5
    config_files = [f for f in os.walk(USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH)]
    config_files_count = len(config_files)
    assert config_files_count == 1
    # ensure fetched InstantFluctuationsEvaluator requirement
    assert "InstantFluctuationsEvaluator.json" in config_files[0][2]
    assert os.path.exists(os.path.join("tentacles", "Evaluator", "RealTime",
                                       "instant_fluctuations_evaluator", "instant_fluctuations_evaluator.py"))
    assert len(config_files[0][2]) == 4


def _enable_loggers():
    set_logging_level([clazz.__name__ for clazz in [InstallWorker, Tentacle]], INFO)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os.path as path


class TentacleType:
    def __init__(self, type_path: str, module_name: str = None):
        self.path: str = type_path
        self.module_name: str = module_name

    @staticmethod
    def from_import_path(root: str, import_path: str):
        """
        Create a tentacle type from a path
        :param root: the root path
        :param import_path: the import path
        :return: the created TentacleType instance
        """
        module_name = import_path.split(".")[-1]
        tentacle_type_path = import_path.split(f"{root}.")[-1]\
            .replace(f".{module_name}", "")\
            .replace(f".", path.sep)
        return TentacleType(tentacle_type_path, module_name)

    def get_last_element(self) -> str:
        """
        :return: the path last element
        """
        return self.path.split(path.sep)[-1]

    def get_root_type(self) -> str:
        """
        :return: the root type
        """
        return self.path.split(path.sep)[0]

    def to_path(self) -> str:
        """
        :return: the tentacle type path
        """
        return self.path

    def __str__(self) -> str:
        return self.path.replace(path.sep, ".")



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os.path as path
import octobot_commons.constants as constants
import octobot_commons.enums as enums

# Default tests tentacles URL
DEFAULT_TENTACLES_URL = "https://www.tentacles.octobot.online/repository/tentacles/officials/base/tests/0.0.2-tests.zip"
DEFAULT_BOT_INSTALL_DIR = "."

# Tentacles files
PYTHON_INIT_FILE = "__init__.py"
CYTHON_HEADER_INIT_FILE = "__init__.pxd"
PYTHON_EXT = ".py"
CYTHON_EXT = ".pxd"
CONFIG_EXT = ".json"
DOCUMENTATION_EXT = ".md"
CONFIG_SCHEMA_EXT = "_schema.json"
TENTACLE_METADATA = "metadata.json"
DEFAULT_TENTACLE_CONFIG = path.join(constants.CONFIG_FOLDER, "default_tentacles_config.json")

# tentacles setup folders back list
FOLDERS_BLACK_LIST = ["__pycache__"]

# Metadata keys
METADATA_VERSION = "version"
METADATA_TENTACLES = "tentacles"
METADATA_ORIGIN_PACKAGE = "origin_package"
METADATA_DEV_MODE = "dev_mode"
METADATA_TENTACLES_REQUIREMENTS = "tentacles-requirements"
METADATA_TENTACLES_GROUP = "tentacles_group"

# Artifact metadata
ARTIFACT_METADATA_FILE = "metadata.yaml"
ARTIFACT_METADATA_VERSION = "version"
ARTIFACT_METADATA_NAME = "name"
ARTIFACT_METADATA_ARTIFACT_TYPE = "type"
ARTIFACT_METADATA_REPOSITORY = "repository"
ARTIFACT_METADATA_AUTHOR = "author"
ARTIFACT_METADATA_ARTIFACTS = "artifacts"
DEFAULT_ARTIFACT_METADATA_AUTHOR = "DrakkarSoftware"

# Requirements
TENTACLE_REQUIREMENT_VERSION_EQUALS = "=="

# Tentacle user config files and folders
USER_REFERENCE_TENTACLE_CONFIG_PATH = path.join(constants.USER_FOLDER, "reference_tentacles_config")
USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH = path.join(USER_REFERENCE_TENTACLE_CONFIG_PATH,
                                                     constants.CONFIG_TENTACLES_FILE)
TENTACLES_SPECIFIC_CONFIG_FOLDER = "specific_config"
USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH = path.join(USER_REFERENCE_TENTACLE_CONFIG_PATH,
                                                         TENTACLES_SPECIFIC_CONFIG_FOLDER)

# Current minimum default tentacles version
TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION = "1.2.0"
DEFAULT_TENTACLES_PACKAGE = "OctoBot-Default-Tentacles"

# Tentacles repository
UNKNOWN_REPOSITORY_LOCATION = "Unknown repository location"

# Tentacles packages
UNKNOWN_TENTACLES_PACKAGE_LOCATION = "Unknown package location"
TENTACLES_PACKAGE_PROFILES_PATH = "profiles"

# exporter
DEFAULT_EXPORT_DIR = "output"

# Artifacts
UNKNOWN_ARTIFACT_VERSION = "unknown_version"

# Tentacles installation folders
TENTACLES_INSTALL_TEMP_DIR = "temp_tentacles"
TENTACLES_REQUIREMENTS_INSTALL_TEMP_DIR = "requirements_temp_tentacles"
TENTACLES_ARCHIVE_ROOT = "reference_tentacles"
TO_REMOVE_FOLDER = "_to_rm"

# Tentacles folders
TENTACLES_PATH = "tentacles"
DEFAULT_BOT_PATH = "."
TENTACLES_BACKTESTING_PATH = "Backtesting"
TENTACLES_EVALUATOR_PATH = "Evaluator"
TENTACLES_INTERFACES_PATH = "Interfaces"
TENTACLES_NOTIFIERS_PATH = "Notifiers"
TENTACLES_SERVICES_PATH = "Services"
TENTACLES_SERVICES_BASES_PATH = "Services_bases"
TENTACLES_SERVICES_FEEDS_PATH = "Services_feeds"
TENTACLES_TRADING_PATH = "Trading"

# Tentacles sub-folders
TENTACLE_MAX_SUB_FOLDERS_LEVEL = 3
TENTACLES_BACKTESTING_COLLECTORS_PATH = "collectors"
TENTACLES_BACKTESTING_CONVERTERS_PATH = "converters"
TENTACLES_BACKTESTING_IMPORTERS_PATH = "importers"
TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH = "exchanges"
TENTACLES_BACKTESTING_THIRD_LEVEL_SOCIAL_PATH = "social"
TENTACLES_EVALUATOR_REALTIME_PATH = "RealTime"
TENTACLES_EVALUATOR_TA_PATH = "TA"
TENTACLES_EVALUATOR_SOCIAL_PATH = "Social"
TENTACLES_EVALUATOR_STRATEGIES_PATH = "Strategies"
TENTACLES_EVALUATOR_UTIL_PATH = "Util"
TENTACLES_TRADING_MODE_PATH = "Mode"
TENTACLES_TRADING_EXCHANGE_PATH = "Exchange"
TENTACLES_WEBSOCKETS_FEEDS_PATH = "feeds"

# Tentacle local module folders
TENTACLE_CONFIG = "config"
TENTACLE_RESOURCES = "resources"
TENTACLE_TESTS = "tests"

# Tentacle creator
TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER = "creator_temp"
TENTACLES_PACKAGE_FORMAT = "zip"
PYTHON_GENERATED_ELEMENTS = {"__pycache__"}
TENTACLES_PACKAGE_IGNORED_ELEMENTS = [".git", ".coveragerc", ".gitignore", ".travis.yml",
                                      "azure-pipelines.yml", "octobot_config.json"]
COMPILED_TENTACLES_TO_REMOVE_FOLDERS = {"build", "temp", "tests"}
PYTHON_GENERATED_ELEMENTS_EXTENSION = {"pyc"}
COMPILED_TENTACLES_TO_KEEP_ELEMENTS = {PYTHON_INIT_FILE}
COMPILED_TENTACLES_TO_REMOVE_ELEMENTS = {".c", ".py", ".pxd", ".pyc", "tests"}
TENTACLE_TEMPLATE_PATH = "templates"
TENTACLE_TEMPLATE_DESCRIPTION = "description"
TENTACLE_CONFIG_TEMPLATE_PRE_EXT = "_config"
TENTACLE_TEMPLATE_EXT = ".template"
TENTACLE_TEMPLATE_PRE_EXT = "_tentacle"
CYTHON_PXD_HEADER = "# cython: language_level=3"
SETUP_FILE = "setup.py"

# Tentacles architecture
TENTACLES_FOLDERS_ARCH = {
    TENTACLES_BACKTESTING_PATH: {
        TENTACLES_BACKTESTING_COLLECTORS_PATH: [
            TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH,
            TENTACLES_BACKTESTING_THIRD_LEVEL_SOCIAL_PATH
        ],
        TENTACLES_BACKTESTING_CONVERTERS_PATH: [
            TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH
        ],
        TENTACLES_BACKTESTING_IMPORTERS_PATH: [
            TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH,
            TENTACLES_BACKTESTING_THIRD_LEVEL_SOCIAL_PATH
        ]
    },
    TENTACLES_EVALUATOR_PATH: [
        TENTACLES_EVALUATOR_REALTIME_PATH,
        TENTACLES_EVALUATOR_SOCIAL_PATH,
        TENTACLES_EVALUATOR_TA_PATH,
        TENTACLES_EVALUATOR_STRATEGIES_PATH,
        TENTACLES_EVALUATOR_UTIL_PATH
    ],
    TENTACLES_SERVICES_PATH: [
        TENTACLES_INTERFACES_PATH,
        TENTACLES_NOTIFIERS_PATH,
        TENTACLES_SERVICES_BASES_PATH,
        TENTACLES_SERVICES_FEEDS_PATH
    ],
    TENTACLES_TRADING_PATH: [
        TENTACLES_TRADING_MODE_PATH,
        TENTACLES_TRADING_EXCHANGE_PATH
    ]
}

TENTACLE_MODULE_FOLDERS = {
    TENTACLE_CONFIG,
    TENTACLE_RESOURCES,
    TENTACLE_TESTS
}


# tentacles files management
TENTACLE_TYPES = [
    TENTACLES_BACKTESTING_PATH,
    TENTACLES_EVALUATOR_PATH,
    TENTACLES_SERVICES_PATH,
    TENTACLES_TRADING_PATH
]

# compiled tentacles paths
PLATFORM_TO_DOWNLOAD_PATH = {
    enums.PlatformsName.WINDOWS: "windows",
    enums.PlatformsName.LINUX: "linux",
    enums.PlatformsName.MAC: "macos",
}



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import os.path as path

import aiofiles

import octobot_tentacles_manager.models.artifact as artifact
import octobot_tentacles_manager.constants as constants


class Tentacle(artifact.Artifact):
    ARTIFACT_NAME = "tentacle"

    def __init__(self, tentacle_root_path, name, tentacle_type):
        super().__init__(name)
        self.tentacle_root_path = tentacle_root_path
        self.tentacle_type = tentacle_type
        self.tentacle_root_type = self.tentacle_type.get_root_type()
        self.tentacle_path = path.join(self.tentacle_root_path, self.tentacle_type.to_path())
        self.tentacle_module_path = None
        self.tentacle_class_names = []
        self.tentacles_requirements = None
        self.tentacle_group = self.name
        self.in_dev_mode = False
        self.metadata = {}

    async def initialize(self) -> None:
        """
        Initialize a tentacle object with its metadata file
        :return: None
        """
        self.tentacle_module_path = path.join(self.tentacle_path, self.name)
        async with aiofiles.open(path.join(self.tentacle_module_path,
                                           constants.TENTACLE_METADATA), "r") as metadata_file:
            self._read_metadata_dict(json.loads(await metadata_file.read()))

    def sync_initialize(self) -> None:
        """
        Initialize synchronously a tentacle object with its metadata file
        :return: None
        """
        try:
            self.tentacle_module_path = path.join(self.tentacle_path, self.name)
            with open(path.join(self.tentacle_path, self.name, constants.TENTACLE_METADATA), "r") as metadata_file:
                self._read_metadata_dict(json.loads(metadata_file.read()))
        except FileNotFoundError:
            pass

    @staticmethod
    def find(iterable, name) -> object:
        """
        Find a tentacle from an iterable
        :param iterable: the iterable
        :param name: the tentacle name
        :return: the tentacle if found else None
        """
        for tentacle in iterable:
            if tentacle.name == name:
                return tentacle
        return None

    def get_simple_tentacle_type(self) -> str:
        """
        :return: the tentacle type from its folder in tentacle files architecture
        """
        return self.tentacle_type.get_last_element()

    def __str__(self) -> str:
        str_rep: str = f"{self.name} {Tentacle.ARTIFACT_NAME} [type: {self.tentacle_type}"
        if self.is_valid():
            return f"{str_rep}, version: {self.version}]"
        else:
            return f"{str_rep}]"

    def extract_tentacle_requirements(self) -> list:
        """
        :return: The tentacle requirement list
        """
        return [self._parse_requirements(component) for component in self.tentacles_requirements]

    def _read_metadata_dict(self, metadata: dict) -> None:
        """
        Load tentacle metadata from its dict
        :param metadata: the tentacle metadata dict
        :return: None
        """
        self.metadata = metadata
        self.version = self.metadata[constants.METADATA_VERSION]
        self.origin_package = self.metadata[constants.METADATA_ORIGIN_PACKAGE]
        self.tentacle_class_names = self.metadata[constants.METADATA_TENTACLES]
        self.tentacles_requirements = self.metadata[constants.METADATA_TENTACLES_REQUIREMENTS]
        # self.tentacle_group is this tentacle name if no provided
        self.tentacle_group = self.metadata.get(constants.METADATA_TENTACLES_GROUP, self.name)
        if constants.METADATA_DEV_MODE in self.metadata:
            self.in_dev_mode = self.metadata[constants.METADATA_DEV_MODE]

    @staticmethod
    def _parse_requirements(requirement) -> list:
        """
        Parse tentacle requirements
        :param requirement: tentacle requirements
        :return: parsed tentacle requirements
        """
        if constants.TENTACLE_REQUIREMENT_VERSION_EQUALS in requirement:
            return requirement.split(constants.TENTACLE_REQUIREMENT_VERSION_EQUALS)
        else:
            return [requirement, None]



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import pytest
from logging import INFO
import os

import octobot_commons.constants as commons_constants
from octobot_commons.logging.logging_util import set_logging_level
from octobot_tentacles_manager.constants import TENTACLES_PATH, TENTACLES_SPECIFIC_CONFIG_FOLDER, \
    USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, DEFAULT_BOT_PATH
from octobot_tentacles_manager.workers.install_worker import InstallWorker

from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.workers.uninstall_worker import UninstallWorker
from octobot_tentacles_manager.util.tentacle_fetching import fetch_and_extract_tentacles
from tests import event_loop, clean, fake_profiles, TEMP_DIR, OTHER_PROFILE

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_uninstall_two_tentacles(clean):
    _enable_loggers()
    tentacles_path = os.path.join("tests", "static", "tentacles.zip")
    await fetch_and_extract_tentacles(TEMP_DIR, tentacles_path, None)
    install_worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    install_worker.tentacles_path_or_url = tentacles_path
    install_worker.tentacles_setup_manager.default_tentacle_config \
        = os.path.join("tests", "static", "default_tentacle_config.json")
    assert await install_worker.process() == 0
    tentacles_files_count = sum(1 for _ in os.walk(TENTACLES_PATH))
    assert tentacles_files_count > 60

    uninstall_worker = UninstallWorker(None, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    uninstall_worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    assert await uninstall_worker.process(["instant_fluctuations_evaluator", "generic_exchange_importer"]) == 0
    tentacles_files_count = sum(1 for _ in os.walk(TENTACLES_PATH))
    assert tentacles_files_count < 60
    with open(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'registered_tentacles': {
                'OctoBot-Default-Tentacles': tentacles_path
            },
            'tentacle_activation': {
                'Backtesting': {},
                'Evaluator': {
                    'OtherInstantFluctuationsEvaluator': False,
                    'OverallStateAnalyser': True,
                    'RedditForumEvaluator': False,
                    'SecondOtherInstantFluctuationsEvaluator': False,
                    'SimpleMixedStrategyEvaluator': True,
                    'TextAnalysis': True
                },
                'Services': {
                    'RedditService': True,
                    'RedditServiceFeed': True
                },
                'Trading': {
                    'DailyTradingMode': True
                }
            }
        }


async def test_profiles_update(clean, fake_profiles):
    _enable_loggers()
    tentacles_path = os.path.join("tests", "static", "tentacles.zip")
    await fetch_and_extract_tentacles(TEMP_DIR, tentacles_path, None)
    install_worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    install_worker.tentacles_path_or_url = tentacles_path
    install_worker.tentacles_setup_manager.default_tentacle_config \
        = os.path.join("tests", "static", "default_tentacle_config.json")
    assert await install_worker.process() == 0
    tentacles_files_count = sum(1 for _ in os.walk(TENTACLES_PATH))
    assert tentacles_files_count > 60

    ref_specific_tentacles_config = os.path.join(os.path.split(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH)[0],
                                                 TENTACLES_SPECIFIC_CONFIG_FOLDER)
    with open(os.path.join(ref_specific_tentacles_config, "InstantFluctuationsEvaluator.json")) as ref_conf:
        instant_fluct_config = json.load(ref_conf)

    uninstall_worker = UninstallWorker(None, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    uninstall_worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    # uninstall 2 tentacles
    assert await uninstall_worker.process(["instant_fluctuations_evaluator", "generic_exchange_importer"]) == 0

    # test tentacles setup config
    with open(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH) as config_f:
        ref_profile_config = json.load(config_f)

        # test profiles tentacles config
        with open(os.path.join(commons_constants.USER_PROFILES_FOLDER,
                               commons_constants.DEFAULT_PROFILE,
                               commons_constants.CONFIG_TENTACLES_FILE)) as default_c:
            assert ref_profile_config == json.load(default_c)
        with open(os.path.join(commons_constants.USER_PROFILES_FOLDER,
                               OTHER_PROFILE,
                               commons_constants.CONFIG_TENTACLES_FILE)) as other_c:
            assert ref_profile_config == json.load(other_c)

    # test specific tentacles config
    default_profile_tentacles_config = os.path.join(commons_constants.USER_PROFILES_FOLDER,
                                                    commons_constants.DEFAULT_PROFILE,
                                                    TENTACLES_SPECIFIC_CONFIG_FOLDER)
    other_profile_tentacles_config = os.path.join(commons_constants.USER_PROFILES_FOLDER,
                                                  OTHER_PROFILE,
                                                  TENTACLES_SPECIFIC_CONFIG_FOLDER)
    for tentacle_config in os.scandir(ref_specific_tentacles_config):
        with open(tentacle_config) as ref_config_file:
            ref_config = json.load(ref_config_file)
        with open(os.path.join(default_profile_tentacles_config, tentacle_config.name)) as default_profile_config_file:
            assert ref_config == json.load(default_profile_config_file)
        with open(os.path.join(other_profile_tentacles_config, tentacle_config.name)) as other_profile_config_file:
            assert ref_config == json.load(other_profile_config_file)

    # however did not remove tentacles specific config files to be able to reuse them later
    with open(os.path.join(default_profile_tentacles_config, "InstantFluctuationsEvaluator.json")) as def_conf:
        assert instant_fluct_config == json.load(def_conf)
    with open(os.path.join(other_profile_tentacles_config, "InstantFluctuationsEvaluator.json")) as other_conf:
        assert instant_fluct_config == json.load(other_conf)


async def test_uninstall_all_tentacles(clean):
    _enable_loggers()
    tentacles_path = os.path.join("tests", "static", "tentacles.zip")
    await fetch_and_extract_tentacles(TEMP_DIR, tentacles_path, None)
    install_worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    install_worker.tentacles_path_or_url = tentacles_path
    install_worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    assert await install_worker.process() == 0
    tentacles_files_count = sum(1 for _ in os.walk(TENTACLES_PATH))
    assert tentacles_files_count > 60

    uninstall_worker = UninstallWorker(None, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    uninstall_worker.tentacles_setup_manager.default_tentacle_config = \
        os.path.join("tests", "static", "default_tentacle_config.json")
    assert await uninstall_worker.process() == 0
    tentacles_files_count = sum(1 for _ in os.walk(TENTACLES_PATH))
    assert tentacles_files_count == 24
    with open(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'registered_tentacles': {},
            'tentacle_activation': {
                'Backtesting': {},
                'Evaluator': {},
                'Services': {},
                'Trading': {}
            }
        }


def _enable_loggers():
    set_logging_level([clazz.__name__ for clazz in [InstallWorker, Tentacle]], INFO)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_tentacles_manager.constants as constants


class Artifact:
    ARTIFACT_NAME = "artifact"

    def __init__(self, name):
        self.name: str = name
        self.version: str = constants.UNKNOWN_ARTIFACT_VERSION
        self.origin_package: str = constants.UNKNOWN_TENTACLES_PACKAGE_LOCATION
        self.origin_repository: str = constants.UNKNOWN_REPOSITORY_LOCATION
        self.output_path: str = None
        self.output_dir: str = None

    def is_valid(self) -> bool:
        """
        :return: True if the artifact version is not a None or a default value
        """
        return self.version is not None and self.version != constants.UNKNOWN_ARTIFACT_VERSION

    def __str__(self):
        str_rep = f"{self.name} {Artifact.ARTIFACT_NAME} ["
        if self.is_valid():
            return f"version: {self.version}]"
        else:
            return f"{str_rep}]"



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import octobot_tentacles_manager.uploaders.nexus_uploader as nexus_uploader


async def upload_file_or_folder_to_nexus(nexus_path: str, artifact_path: str, artifact_alias: str = None) -> int:
    if os.path.isfile(artifact_path):
        return await upload_file_to_nexus(nexus_path=nexus_path,
                                          file_path=artifact_path,
                                          alias_file_name=artifact_alias)
    elif os.path.isdir(artifact_path):
        return await upload_folder_to_nexus(nexus_path=nexus_path,
                                            folder_path=artifact_path,
                                            alias_folder_name=artifact_alias)


async def upload_file_to_nexus(nexus_path: str, file_path: str, alias_file_name: str = None) -> int:
    uploader: nexus_uploader.NexusUploader = nexus_uploader.NexusUploader()
    upload_result: int = await uploader.upload_file(upload_path=nexus_path,
                                                    file_path=file_path,
                                                    destination_file_name=alias_file_name)
    await uploader.close_session()
    return upload_result


async def upload_folder_to_nexus(nexus_path: str, folder_path: str, alias_folder_name: str = None) -> int:
    uploader: nexus_uploader.NexusUploader = nexus_uploader.NexusUploader()
    upload_result: int = await uploader.upload_folder(upload_path=nexus_path,
                                                      folder_path=folder_path,
                                                      destination_folder_name=alias_folder_name)
    await uploader.close_session()
    return upload_result



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import os
from os import path, remove, mkdir, scandir, walk
from os.path import join, getsize
from shutil import rmtree

import pytest

from tests.api import install_tentacles, TENTACLE_PACKAGE, TEST_EXPORT_DIR
from octobot_tentacles_manager.api.creator import create_tentacles_package, create_all_tentacles_bundle
from octobot_tentacles_manager.constants import TENTACLES_PATH, PYTHON_INIT_FILE, TENTACLES_EVALUATOR_PATH, \
    TENTACLES_EVALUATOR_REALTIME_PATH, TENTACLE_METADATA, METADATA_VERSION, METADATA_ORIGIN_PACKAGE, \
    METADATA_TENTACLES, METADATA_TENTACLES_REQUIREMENTS, METADATA_DEV_MODE, TENTACLES_TRADING_PATH, \
    TENTACLES_TRADING_MODE_PATH, TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER, TENTACLES_SERVICES_PATH, \
    TENTACLES_BACKTESTING_IMPORTERS_PATH, TENTACLES_BACKTESTING_PATH, TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_create_folder_tentacles_package(install_tentacles):
    # set instant_fluctuations_evaluator in dev mode
    tentacle_path = path.join(TENTACLES_PATH, TENTACLES_EVALUATOR_PATH, TENTACLES_EVALUATOR_REALTIME_PATH)
    with open(path.join(tentacle_path, "instant_fluctuations_evaluator", TENTACLE_METADATA), "w+") as metadata:
        new_metadata = {
            METADATA_VERSION: "1.2.0",
            METADATA_ORIGIN_PACKAGE: "OctoBot-Default-Tentacles",
            METADATA_TENTACLES: ["InstantFluctuationsEvaluator"],
            METADATA_TENTACLES_REQUIREMENTS: [],
            METADATA_DEV_MODE: True
        }
        json.dump(new_metadata, metadata)

    # add generated python file
    random_content = "123"
    generated_file_path = path.join(TENTACLES_PATH, TENTACLES_TRADING_PATH, "file.pyc")
    with open(generated_file_path, "w+") as generated_file:
        generated_file.write(random_content)
    # add generated python folder
    generated_folder_path = path.join(TENTACLES_PATH, TENTACLES_TRADING_PATH, "__pycache__")
    mkdir(generated_folder_path)

    # add gitignore file that should not be copied
    with open(path.join(TENTACLES_PATH, ".gitignore"), "w+") as ignored_file:
        ignored_file.write(random_content)

    # create folder to force folder merge
    mkdir(TENTACLE_PACKAGE)

    with open(path.join(TENTACLE_PACKAGE, "not_tentacle_file"), "w+") as rand_file:
        rand_file.write(random_content)

    mkdir(path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH))

    with open(path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH, "rnd"), "w+") as rand_file:
        rand_file.write(random_content)

    assert await create_tentacles_package(TENTACLE_PACKAGE, output_dir=".", in_zip=False) == 0
    assert path.exists(TENTACLE_PACKAGE)

    # random init file still here
    with open(path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH, "rnd")) as rand_file:
        assert random_content == rand_file.read()

    # file that is not in tentacles arch is copied since it's not in a zip
    assert path.exists(path.join(TENTACLE_PACKAGE, "not_tentacle_file"))

    # ignored file is not copied
    assert not path.exists(path.join(TENTACLE_PACKAGE, ".gitignore"))

    # generated file not copied
    assert not path.exists(path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH, "file.pyc"))

    # generated folder not copied
    assert not path.exists(path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH, "__pycache__"))

    # did not add instant_fluctuations_evaluator (in dev-mode)
    package_tentacle_path = path.join(TENTACLE_PACKAGE, TENTACLES_EVALUATOR_PATH, TENTACLES_EVALUATOR_REALTIME_PATH)
    assert not path.exists(path.join(package_tentacle_path, "instant_fluctuations_evaluator"))

    # added other_instant_fluctuations_evaluator
    assert path.exists(path.join(package_tentacle_path, "other_instant_fluctuations_evaluator"))

    # added daily_trading_mode
    trading_mode_tentacle_path = path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH, TENTACLES_TRADING_MODE_PATH)
    assert path.exists(path.join(trading_mode_tentacle_path, "daily_trading_mode"))
    assert path.exists(path.join(trading_mode_tentacle_path, "daily_trading_mode", "daily_trading_mode.py"))

    # removed python init files
    assert not path.exists(path.join(TENTACLE_PACKAGE, PYTHON_INIT_FILE))
    evaluator_path = path.join(TENTACLE_PACKAGE, TENTACLES_EVALUATOR_PATH)
    assert not path.exists(path.join(evaluator_path, PYTHON_INIT_FILE))
    assert not path.exists(path.join(evaluator_path, TENTACLES_EVALUATOR_REALTIME_PATH, PYTHON_INIT_FILE))

    # add every not in dev tentacles
    python_files = tuple(e[2] for e in walk(TENTACLE_PACKAGE) if any(i for i in e[2] if i.endswith(".py")))
    assert len(python_files) == 10


async def test_create_folder_tentacles_package_with_package_selector(install_tentacles):
    tentacle_path = path.join(TENTACLES_PATH, TENTACLES_TRADING_PATH, TENTACLES_TRADING_MODE_PATH)
    with open(path.join(tentacle_path, "daily_trading_mode", TENTACLE_METADATA), "w+") as metadata:
        new_metadata = {
            METADATA_VERSION: "1.2.0",
            METADATA_ORIGIN_PACKAGE: "OctoBot-Not-Quite-Default-Tentacles",
            METADATA_TENTACLES: ["DailyTradingMode"],
            METADATA_TENTACLES_REQUIREMENTS: [],
            METADATA_DEV_MODE: False
        }
        json.dump(new_metadata, metadata)

    assert await create_tentacles_package(TENTACLE_PACKAGE, output_dir=".", in_zip=False,
                                          exported_tentacles_package="OctoBot-Not-Quite-Default-Tentacles") == 0
    assert path.exists(TENTACLE_PACKAGE)

    # added daily_trading_mode
    trading_mode_tentacle_path = path.join(TENTACLE_PACKAGE, TENTACLES_TRADING_PATH, TENTACLES_TRADING_MODE_PATH)
    assert path.exists(path.join(trading_mode_tentacle_path, "daily_trading_mode"))
    assert path.exists(path.join(trading_mode_tentacle_path, "daily_trading_mode", "daily_trading_mode.py"))

    # did not add anything else then daily trading mode files and tests
    python_files = tuple(e[2] for e in walk(TENTACLE_PACKAGE) if any(i for i in e[2] if i.endswith(".py")))
    assert len(python_files) == 2


async def test_create_zipped_tentacles_package(install_tentacles):
    tentacle_package = "tentacle_package.zip"
    assert await create_tentacles_package(tentacle_package, output_dir=".") == 0
    assert path.exists(tentacle_package)
    assert not path.exists(TENTACLES_PACKAGE_CREATOR_TEMP_FOLDER)
    remove(tentacle_package)


async def test_create_cythonized_tentacles_package(install_tentacles):
    speedup_cythonization()
    exchange_importer_path = join(TENTACLES_PATH,
                                  TENTACLES_BACKTESTING_PATH,
                                  TENTACLES_BACKTESTING_IMPORTERS_PATH,
                                  TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH,
                                  "generic_exchange_importer")
    # create artificial sub folder python file to ensure it will also get compiled
    mkdir(join(exchange_importer_path, "plop"))
    with open(join(exchange_importer_path, "plop", "file.py"), "w+") as file_w:
        file_w.write("plop=0")
    with open(join(exchange_importer_path, "plop", PYTHON_INIT_FILE), "w+") as file_w:
        file_w.write("")
    assert await create_tentacles_package(TENTACLE_PACKAGE, output_dir=".", in_zip=False, cythonize=True) == 0
    _check_compiled_tentacle(join(TENTACLE_PACKAGE,
                                  TENTACLES_BACKTESTING_PATH,
                                  TENTACLES_BACKTESTING_IMPORTERS_PATH,
                                  TENTACLES_BACKTESTING_THIRD_LEVEL_EXCHANGES_PATH,
                                  "generic_exchange_importer"),
                             3)
    _check_compiled_tentacle(join(TENTACLE_PACKAGE,
                                  TENTACLES_TRADING_PATH,
                                  TENTACLES_TRADING_MODE_PATH,
                                  "daily_trading_mode"),
                             2)


async def test_create_all_tentacles_bundle_cleaned_not_zipped_but_cythonized(install_tentacles):
    speedup_cythonization()
    assert await create_all_tentacles_bundle(TEST_EXPORT_DIR,
                                             in_zip=False,
                                             cythonize=True,
                                             should_remove_artifacts_after_use=True) == 0

    # test compiled content
    _check_compiled_tentacle(join(TEST_EXPORT_DIR, "generic_exchange_importer@1.2.0"), 2)
    _check_compiled_tentacle(join(TEST_EXPORT_DIR, "daily_trading_mode@1.2.0"), 2)

    # test should_remove_artifacts_after_use
    assert not os.path.isfile(os.path.join(TEST_EXPORT_DIR, "generic_exchange_importer"))
    assert not os.path.isfile(os.path.join(TEST_EXPORT_DIR, "daily_trading_mode"))

    # test bundle metadata
    assert_directory_has_file_with_content(os.path.join(TEST_EXPORT_DIR, "generic_exchange_importer@1.2.0"),
                                           "metadata.yaml",
                                           "author: DrakkarSoftware\nname: generic_exchange_importer\nrepository: Unknown repository location\ntype: tentacle\nversion: 1.2.0\n")


async def test_create_all_tentacles_bundle_zipped_not_cleaned_and_zipped(install_tentacles):
    assert await create_all_tentacles_bundle(TEST_EXPORT_DIR,
                                             in_zip=True,
                                             cythonize=False,
                                             should_remove_artifacts_after_use=False,
                                             should_zip_bundle=True) == 0

    # test not should_remove_artifacts_after_use
    assert os.path.isfile(os.path.join(TEST_EXPORT_DIR, "generic_exchange_importer.zip"))
    assert os.path.isfile(os.path.join(TEST_EXPORT_DIR, "daily_trading_mode.zip"))
    assert os.path.isfile(os.path.join(TEST_EXPORT_DIR, "generic_exchange_importer_1.2.0.zip"))
    assert os.path.isfile(os.path.join(TEST_EXPORT_DIR, "daily_trading_mode_1.2.0.zip"))
    assert not os.path.isdir(os.path.join(TEST_EXPORT_DIR, "generic_exchange_importer@1.2.0"))
    assert not os.path.isdir(os.path.join(TEST_EXPORT_DIR, "daily_trading_mode@1.2.0"))


async def test_create_all_tentacles_bundle_not_zipped_not_cleaned_and_zipped(install_tentacles):
    assert await create_all_tentacles_bundle(TEST_EXPORT_DIR,
                                             in_zip=True,
                                             cythonize=False,
                                             should_remove_artifacts_after_use=False,
                                             should_zip_bundle=False) == 0

    # test not should_remove_artifacts_after_use
    assert os.path.isfile(os.path.join(TEST_EXPORT_DIR, "generic_exchange_importer.zip"))
    assert os.path.isfile(os.path.join(TEST_EXPORT_DIR, "daily_trading_mode.zip"))
    assert not os.path.isfile(os.path.join(TEST_EXPORT_DIR, "generic_exchange_importer_1.2.0.zip"))
    assert not os.path.isfile(os.path.join(TEST_EXPORT_DIR, "daily_trading_mode_1.2.0.zip"))
    assert os.path.isdir(os.path.join(TEST_EXPORT_DIR, "generic_exchange_importer@1.2.0"))
    assert os.path.isdir(os.path.join(TEST_EXPORT_DIR, "daily_trading_mode@1.2.0"))


def assert_directory_has_file_with_content(directory_to_check, expected_file, expected_content):
    with open(os.path.join(directory_to_check, expected_file), "r") as file_to_test:
        assert file_to_test.read() == expected_content


def speedup_cythonization():
    # remove evaluator and services tentacles to speed up compilation
    for folder in (TENTACLES_EVALUATOR_PATH, TENTACLES_SERVICES_PATH):
        rmtree(join(TENTACLES_PATH, folder))

def _check_compiled_tentacle(tentacle_path, expected_dir_count):
    dir_count = 0
    has_compiled_file = False
    for entry in scandir(tentacle_path):
        if entry.is_file():
            if entry.name.endswith(".py"):
                assert entry.name == PYTHON_INIT_FILE
            elif not (entry.name.endswith(".json") or entry.name.endswith(".yaml")):
                assert entry.name.endswith(".pyd") or entry.name.endswith(".so")
                assert getsize(entry) >= 100
                has_compiled_file = True
        elif entry.is_dir():
            dir_count += 1
    assert has_compiled_file
    assert dir_count == expected_dir_count



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import pytest
from logging import INFO
from shutil import rmtree
from os import walk, path

from octobot_commons.logging.logging_util import set_logging_level
from octobot_tentacles_manager.constants import USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH, \
    USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, TENTACLES_PATH, DEFAULT_BOT_PATH, UNKNOWN_TENTACLES_PACKAGE_LOCATION
from octobot_tentacles_manager.workers.install_worker import InstallWorker
from octobot_tentacles_manager.models.tentacle_factory import TentacleFactory
from octobot_tentacles_manager.workers.update_worker import UpdateWorker
from octobot_tentacles_manager.models.tentacle import Tentacle
from octobot_tentacles_manager.util.tentacle_fetching import fetch_and_extract_tentacles
from tests import event_loop, clean, TEMP_DIR

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_update_two_tentacles(clean):
    _enable_loggers()
    await fetch_and_extract_tentacles(TEMP_DIR, path.join("tests", "static", "tentacles.zip"), None)
    install_worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    install_worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    await install_worker.process(["instant_fluctuations_evaluator",
                                  "generic_exchange_importer",
                                  "text_analysis"])
    rmtree(TEMP_DIR)

    # edit instant_fluctuations_evaluator config to ensure file is not replaced
    config_path = path.join(USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH, "InstantFluctuationsEvaluator.json")
    with open(config_path, "r+") as config_f:
        new_content = f"{config_f.read()},"
        config_f.write(",")
    await fetch_and_extract_tentacles(TEMP_DIR, path.join("tests", "static", "update_tentacles.zip"), None)
    update_worker = UpdateWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    update_worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    assert await update_worker.process(["instant_fluctuations_evaluator", "generic_exchange_importer"]) == 0

    # ensure instant_fluctuations_evaluator file is not replaced
    with open(config_path, "r") as config_f:
        assert new_content == config_f.read()

    # test installed files
    trading_mode_files_count = sum(1 for _ in walk(path.join(TENTACLES_PATH, "Trading", "Mode")))
    assert trading_mode_files_count == 1
    backtesting_mode_files_count = sum(1 for _ in walk(path.join(TENTACLES_PATH, "Backtesting", "importers")))
    assert backtesting_mode_files_count == 7
    config_files = [f for f in walk(USER_REFERENCE_TENTACLE_SPECIFIC_CONFIG_PATH)]
    config_files_count = len(config_files)
    assert config_files_count == 1
    assert "InstantFluctuationsEvaluator.json" in config_files[0][2]
    assert "DailyTradingMode.json" not in config_files[0][2]
    assert len(config_files[0][2]) == 1

    # test tentacles config
    with open(USER_REFERENCE_TENTACLE_CONFIG_FILE_PATH, "r") as config_f:
        assert json.load(config_f) == {
            'registered_tentacles': {
                'OctoBot-Default-Tentacles': UNKNOWN_TENTACLES_PACKAGE_LOCATION
            },
            'tentacle_activation': {
                'Backtesting': {
                    'GenericExchangeDataImporter': True
                },
                'Evaluator': {
                    'InstantFluctuationsEvaluator': True,
                    'TextAnalysis': True
                }
            }
        }

    # check updated versions
    factory = TentacleFactory("tentacles")
    from tentacles.Evaluator.RealTime import instant_fluctuations_evaluator
    ife_tentacle_data = await factory.create_and_load_tentacle_from_module(instant_fluctuations_evaluator)
    assert ife_tentacle_data.version == "1.3.0"
    # not updated because update version is "1.1.9"
    import tentacles.Backtesting.importers.exchanges.generic_exchange_importer as gei
    gei_tentacle_data = await factory.create_and_load_tentacle_from_module(gei)
    assert gei_tentacle_data.version == "1.2.0"
    import tentacles.Evaluator.Util.text_analysis
    ta_tentacle_data = await factory.create_and_load_tentacle_from_module(tentacles.Evaluator.Util.text_analysis)
    assert ta_tentacle_data.version == "1.2.0"


async def test_update_all_tentacles(clean):
    _enable_loggers()
    await fetch_and_extract_tentacles(TEMP_DIR, path.join("tests", "static", "tentacles.zip"), None)
    install_worker = InstallWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    install_worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    await install_worker.process()
    rmtree(TEMP_DIR)
    await fetch_and_extract_tentacles(TEMP_DIR, path.join("tests", "static", "update_tentacles.zip"), None)
    update_worker = UpdateWorker(TEMP_DIR, TENTACLES_PATH, DEFAULT_BOT_PATH, False, None)
    update_worker.tentacles_setup_manager.default_tentacle_config = \
        path.join("tests", "static", "default_tentacle_config.json")
    assert await update_worker.process() == 0

    # check updated versions
    factory = TentacleFactory(TENTACLES_PATH)
    from tentacles.Evaluator.RealTime import instant_fluctuations_evaluator
    ife_tentacle_data = await factory.create_and_load_tentacle_from_module(instant_fluctuations_evaluator)
    assert ife_tentacle_data.version == "1.3.0"
    import tentacles.Backtesting.importers.exchanges.generic_exchange_importer as gei
    gei_tentacle_data = await factory.create_and_load_tentacle_from_module(gei)
    # not updated because update version is "1.1.9"
    assert gei_tentacle_data.version == "1.2.0"
    import tentacles.Evaluator.Util.text_analysis
    ta_tentacle_data = await factory.create_and_load_tentacle_from_module(tentacles.Evaluator.Util.text_analysis)
    assert ta_tentacle_data.version == "1.3.0"
    import tentacles.Trading.Mode.daily_trading_mode as dtm
    dtm_tentacle_data = await factory.create_and_load_tentacle_from_module(dtm)
    assert dtm_tentacle_data.version == "1.3.0"


def _enable_loggers():
    set_logging_level([clazz.__name__ for clazz in [InstallWorker, Tentacle]], INFO)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import os
import shutil

import octobot_tentacles_manager.constants as constants


def remove_unnecessary_files(directory) -> None:
    """
    Remove unnecessary files in directory from
    PYTHON_GENERATED_ELEMENTS and PYTHON_GENERATED_ELEMENTS_EXTENSION list
    :param directory: the directory to clean
    :return: None
    """
    for element in os.scandir(directory):
        element_ext = element.name.split(".")[-1]
        if element.name in constants.PYTHON_GENERATED_ELEMENTS or \
                (element_ext in constants.PYTHON_GENERATED_ELEMENTS_EXTENSION and element.is_file()):
            remove_dir_or_file(element)
        elif element.is_dir():
            remove_unnecessary_files(element)


def remove_dir_or_file(element_path: os.path) -> None:
    """
    Remove an os.path element
    :param element_path: the element to remove
    :return: None
    """
    if element_path.is_dir():
        shutil.rmtree(element_path)
    elif element_path.is_file():
        os.remove(element_path)


def remove_dir_or_file_from_path(element_path: str) -> None:
    """
    Remove an os.path element
    :param element_path: the element path to remove
    :return: None
    """
    if os.path.isdir(element_path):
        shutil.rmtree(element_path)
    elif os.path.isfile(element_path):
        os.remove(element_path)

def remove_non_tentacles_files(directory, logger) -> None:
    """
    Cleanup a directory from non tentacle files
    :param directory: the directory path to clean
    :param logger: the logger to use when an error happened
    :return: None
    """
    for element in os.scandir(directory):
        if element.name not in set(constants.TENTACLES_FOLDERS_ARCH):
            try:
                remove_dir_or_file(element)
            except Exception as e:
                logger.error(f"Error when cleaning up temporary folder: {e}")



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import os

import octobot_tentacles_manager.exporters.artifact_exporter as artifact_exporter
import octobot_tentacles_manager.models as models
import octobot_tentacles_manager.constants as constants


class TentacleExporter(artifact_exporter.ArtifactExporter):
    def __init__(self,
                 artifact: models.Tentacle,
                 tentacles_folder: str,
                 output_dir: str = constants.DEFAULT_EXPORT_DIR,
                 should_cythonize: bool = False,
                 should_zip: bool = False,
                 with_dev_mode: bool = False):
        super().__init__(artifact,
                         tentacles_folder=tentacles_folder,
                         output_dir=output_dir,
                         should_cythonize=should_cythonize,
                         should_zip=should_zip,
                         with_dev_mode=with_dev_mode)

    async def prepare_export(self):
        if not os.path.exists(self.working_folder):
            os.makedirs(self.working_folder)

        if self.should_zip:
            self.copy_directory_content_to_temporary_dir(self.artifact.tentacle_module_path)
        else:
            self.copy_directory_content_to_working_dir(self.artifact.tentacle_module_path)

    async def after_export(self):
        pass



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import abc
import octobot_commons.logging as logging


class Uploader:
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.logger = logging.get_logger(self.__class__.__name__)

    @abc.abstractmethod
    async def upload_file(self, upload_path: str, file_path: str):
        raise NotImplementedError("upload_file is not implemented")

    @abc.abstractmethod
    async def upload_folder(self, upload_path: str, folder_path: str):
        raise NotImplementedError("upload_folder is not implemented")


#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_tentacles_manager.models.artifact as artifact


class TentaclePackage(artifact.Artifact):
    ARTIFACT_NAME = "tentacle package"
    ARTIFACT_SUFFIX = "package"
    ARTIFACT_VERSION_SEPARATOR = "@"

    def __init__(self, name=None):
        super().__init__(name)
        self.artifacts: list = []

    def add_artifact(self, new_artifact: artifact.Artifact) -> None:
        """
        Add a new artifact to the package
        Used to manage the artifact name :
        - If only one artifact is present in the package -> use its name and version
        - If more artifacts are present, use the first artifact's origin package as name
        :param new_artifact: the artifact to add
        :return: None
        """
        self.artifacts.append(new_artifact)
        if len(self.artifacts) == 1:
            self.name = f"{self.artifacts[0].name}{TentaclePackage.ARTIFACT_VERSION_SEPARATOR}" \
                        f"{self.artifacts[0].version}"
        elif len(self.artifacts) > 1:
            self.name = f"{self.artifacts[0].origin_package}{TentaclePackage.ARTIFACT_VERSION_SEPARATOR}" \
                        f"{self.version}_{self.ARTIFACT_SUFFIX}"

    def __str__(self):
        str_rep = f"{self.name} {TentaclePackage.ARTIFACT_NAME} ["
        if self.is_valid():
            return f"version: {self.version}]"
        else:
            return f"{str_rep}]"
